from math import ceil, log
import numpy as np
from pydtnn.utils.constants import NetworkAlgEnum


def roofline(intensity, cpu_speed, memory_bw):
    # print ("COMPUTE_BOUND") if (cpu_speed < memory_bw * intens) else print ("MEMORY_BOUND")
    return min(cpu_speed, memory_bw * intensity)


def flops2time(flops: int, memops: int, cpu_speed: float, memory_bw: float, dtype: type | np.dtype) -> np.ndarray[np.float32]:
    bfp = np.dtype(dtype).itemsize
    speed = roofline(flops / (bfp * memops), cpu_speed, memory_bw)
    time = flops / (speed + 1e-8)
    comp_time = flops / (cpu_speed + 1e-8)
    return np.array([time, comp_time, time - comp_time, 0], dtype=np.float32)


def im2col_time(m: int, n: int, cpu_speed: float, memory_bw: float, dtype: type | np.dtype) -> np.ndarray[np.float32]:
    flops, memops = (0, m * n)
    return flops2time(flops, memops, cpu_speed, memory_bw, dtype)


def col2im_time(m: int, n: int, cpu_speed: float, memory_bw: float, dtype: type | np.dtype) -> np.ndarray[np.float32]:
    flops, memops = (m * n, m * n)
    return flops2time(flops, memops, cpu_speed, memory_bw, dtype)


def matmul_time(m: int, n: int, k: int, cpu_speed: float, memory_bw: float, dtype: type | np.dtype) -> np.ndarray[np.float32]:
    flops, memops = (2.0 * m * n * k, m * n + m * k + n * k)
    return flops2time(flops, memops, cpu_speed, memory_bw, dtype)


def allreduce_time(elems: int, cpu_speed: float, network_bw: float, network_lat: float,
                   network_alg: str, nprocs: int, dtype: type | np.dtype) -> np.ndarray[np.float32]:
    bfp = np.dtype(dtype).itemsize
    time = 0
    match network_alg:
        case NetworkAlgEnum.BTA:
            time = 2.0 * log(nprocs, 2) * network_lat + \
                2.0 * ((nprocs - 1.0) / nprocs) * ((elems * bfp * 8.0) / network_bw) + \
                ((nprocs - 1.0) / nprocs) * (elems / cpu_speed)
            comp_time = ((nprocs - 1.0) / nprocs) * (elems / cpu_speed)
        case NetworkAlgEnum.VDG:
            time = ceil(log(nprocs, 2)) * network_lat + \
                2.0 * ceil(log(nprocs, 2)) * ((elems * bfp * 8.0) / network_bw) + \
                ceil(log(nprocs, 2)) * (elems / cpu_speed)
            comp_time = ceil(log(nprocs, 2)) * (elems / cpu_speed)
        case _ :
            raise ValueError(f"network_alg ({network_alg}) not in {list(NetworkAlgEnum)}")
    # print("allreduce_time; s; %8d; t; %8.8f" % (elems, time))
    return np.array([time, 0, 0, time], dtype=np.float32)


def scatter_time(elems: int, cpu_speed: float, network_bw: float, network_lat: float,
                 network_alg: str, nprocs: int, dtype: type | np.dtype) -> np.ndarray[np.float32]:
    bfp = np.dtype(dtype).itemsize
    time = 0
    match network_alg:
        case NetworkAlgEnum.BTA:
            time = ceil(log(nprocs, 2)) * network_lat + \
                ((nprocs - 1) / nprocs) * ((elems * bfp * 8.0) / network_bw)
        case NetworkAlgEnum.VDG:
            time = log(nprocs, 2) * network_lat + \
                ((nprocs - 1) / nprocs) * ((elems * bfp * 8.0) / network_bw)
        case _ :
            raise ValueError(f"network_alg ({network_alg}) not in {list(NetworkAlgEnum)}")
    # print("scatter_time; s; %8d; t; %8.8f" % (elems, time))
    return np.array([time, 0, 0, time], dtype=np.float32)


def reduce_time(elems: int, cpu_speed: float, network_bw: float, network_lat: float,
                network_alg: str, nprocs: int, dtype: type | np.dtype) -> np.ndarray[np.float32]:
    bfp = np.dtype(dtype).itemsize
    time, comp_time = 0, 0
    match network_alg:
        case NetworkAlgEnum.BTA:
            comp_time = ceil(log(nprocs, 2)) * (elems / cpu_speed)
            time = ceil(log(nprocs, 2)) * network_lat + \
                ceil(log(nprocs, 2)) * (elems * bfp * 8.0) / network_bw + \
                comp_time
        case NetworkAlgEnum.VDG:
            comp_time = ((nprocs - 1.0) / nprocs) * (elems / cpu_speed)
            time = 2.0 * log(nprocs, 2) * network_lat + \
                2.0 * ((nprocs - 1.0) / nprocs) * ((elems * bfp * 8.0) / network_bw) + \
                comp_time        
        case _ :
            raise ValueError(f"network_alg ({network_alg}) not in {list(NetworkAlgEnum)}")

    # print("reduce_time; s; %8d; t; %8.8f" % (elems, time))
    return np.array([time, comp_time, 0, time - comp_time], dtype=np.float32)


def bcast_time(elems: int, cpu_speed: float, network_bw: float, network_lat: float,
               network_alg: str, nprocs: int, dtype: type | np.dtype) -> np.ndarray[np.float32]:
    bfp = np.dtype(dtype).itemsize
    time = 0
    match network_alg:
        case NetworkAlgEnum.BTA:
            time = ceil(log(nprocs, 2)) * ((3 * network_lat) +
                                        ((elems * bfp * 8.0) / network_bw))
        case NetworkAlgEnum.VDG:
            time = (log(nprocs, 2) + nprocs - 1.0) * (network_lat) + \
                2.0 * ((nprocs - 1.0) / nprocs) * ((elems * bfp * 8.0) / network_bw)
        # print("bcast_time; s; %8d; t; %8.8f" % (elems, time))
        case _ :
            raise ValueError(f"network_alg ({network_alg}) not in {list(NetworkAlgEnum)}")

    return np.array([time, 0, 0, time], dtype=np.float32)


def scatter_time(elems: int, cpu_speed: float, network_bw: float, network_lat: float,
                 network_alg: str, nprocs: int, dtype: type | np.dtype) -> np.ndarray[np.float32]:
    bfp = np.dtype(dtype).itemsize
    time = 0
    match network_alg:
        case NetworkAlgEnum.BTA:
            time = ceil(log(nprocs, 2)) * network_lat + \
                (((nprocs - 1) / nprocs)) * ((elems * bfp * 8.0) / network_bw)
        case NetworkAlgEnum.VDG:
            time = log(nprocs) * (network_lat) + \
                (((nprocs - 1) / nprocs)) * ((elems * bfp * 8.0) / network_bw)
        case _ :
            raise ValueError(f"network_alg ({network_alg}) not in {list(NetworkAlgEnum)}")
    # print("scatter_time; s; %8d; t; %8.8f" % (elems, time))
    return time


def gather_time(elems: int, cpu_speed: float, network_bw: float, network_lat: float,
                network_alg: str, nprocs: int, dtype: type | np.dtype) -> np.ndarray[np.float32]:
    time = bcast_time(elems, cpu_speed, network_bw, network_lat, network_alg, nprocs, dtype)
    # print("gather_time; s; %8d; t; %8.8f" % (elems, time))
    return time


def allgather_time(elems: int, cpu_speed: float, network_bw: float, network_lat: float,
                   network_alg: str, nprocs: int, dtype: type | np.dtype) -> np.ndarray[np.float32]:
    bfp = np.dtype(dtype).itemsize
    time = 0
    match network_alg:
        case NetworkAlgEnum.BTA:
            time = (nprocs - 1) * network_lat + \
                ((nprocs - 1) / nprocs) * ((elems * bfp * 8.0) / network_bw)
        case NetworkAlgEnum.VDG:
            time = ceil(log(nprocs, 2)) * (4 * network_lat) + \
                (((nprocs - 1) / nprocs) * ((elems * bfp * 8.0) / network_bw))
        case _ :
            raise ValueError(f"network_alg ({network_alg}) not in {list(NetworkAlgEnum)}")
    # print("allgather_time; s; %8d; t; %8.8f" % (elems, time))
    return np.array([time, 0, 0, time], dtype=np.float32)


def reduce_scatter_time(elems: int, cpu_speed: float, network_bw: float, network_lat: float,
                        network_alg: str, nprocs: int, dtype: type | np.dtype) -> np.ndarray[np.float32]:
    bfp = np.dtype(dtype).itemsize
    time = 0
    match network_alg:
        case NetworkAlgEnum.BTA:
            comp_time = ((nprocs - 1) / nprocs) * (elems / cpu_speed)
            time = (nprocs - 1) * network_lat + \
                ((nprocs - 1) / nprocs) * ((elems * bfp * 8.0) / network_bw) + \
                comp_time
        case NetworkAlgEnum.VDG:
            comp_time = ((nprocs - 1) / nprocs) * (elems / cpu_speed)
            time = (nprocs - 1) * network_lat + \
                ((nprocs - 1) / nprocs) * ((elems * bfp * 8.0) / network_bw) + \
                comp_time
        case _ :
            raise ValueError(f"network_alg ({network_alg}) not in {list(NetworkAlgEnum)}")
        # print("reduce_scatter_time; s; %8d; t; %8.8f" % (elems, time))
    return np.array([time, comp_time, 0, time - comp_time], dtype=np.float32)



import ctypes
import copy
from enum import StrEnum, auto

import numpy as np

from pydtnn.utils.constants import ArrayShape
from pydtnn.utils.tensor import TensorFormat, encode_shape, decode_shape

try:
    import pycuda.gpuarray as gpuarray  # type: ignore
    from pycuda import driver as pycuda_driver  # type: ignore
    from pydtnn.libs import libcudnn as cudnn
except Exception:
    pass


class TensorGPU:

    class TensorTypeEnum(StrEnum):
        TENSOR = auto()
        FILTER = auto()
        SEQ = auto()
        OTHER = auto()
    # ---  END EnumTensorType --- #

    @staticmethod
    def create_empty_tensor(shape: ArrayShape, dtype: np.dtype,
                            tensor_format: TensorFormat, cudnn_dtype: int,
                            tensor_type: TensorTypeEnum = TensorTypeEnum.TENSOR, desc: int | None = None,
                            gpudirect: bool = False, cublas: bool = False):
        gpu_arr = gpuarray.empty(shape, dtype)
        return TensorGPU(gpu_arr=gpu_arr, tensor_format=tensor_format, cudnn_dtype=cudnn_dtype,
                         tensor_type=tensor_type, desc=desc, gpudirect=gpudirect, cublas=cublas)
    # ---

    @staticmethod
    def create_zeros_tensor(shape: ArrayShape, dtype: np.dtype,
                            tensor_format: TensorFormat, cudnn_dtype: int,
                            tensor_type: TensorTypeEnum = TensorTypeEnum.TENSOR, desc: int | None = None,
                            gpudirect: bool = False, cublas: bool = False):
        gpu_arr = gpuarray.zeros(shape, dtype)
        return TensorGPU(gpu_arr=gpu_arr, tensor_format=tensor_format, cudnn_dtype=cudnn_dtype,
                         tensor_type=tensor_type, desc=desc, gpudirect=gpudirect, cublas=cublas)
    
    @staticmethod
    def initialize_gpu_direct(drv: "pycuda_driver", shape: ArrayShape, dtype: np.dtype,
                              tensor_format: TensorFormat, cudnn_dtype: int,
                              tensor_type: TensorTypeEnum = TensorTypeEnum.TENSOR,
                              desc: int | None = None, gpudirect: bool = False, cublas: bool = False) -> tuple[np.ndarray, "TensorGPU"]:
        x_cpu = drv.aligned_zeros(shape, dtype)
        x_gpu = drv.register_host_memory(x_cpu, flags=drv.mem_host_register_flags.DEVICEMAP)

        x_gpu = TensorGPU(x_gpu, tensor_format=tensor_format, cudnn_dtype=cudnn_dtype, tensor_type=tensor_type,
                          desc=desc, gpudirect=gpudirect, cublas=cublas)

        return (x_cpu, x_gpu)
    # ---

    @staticmethod
    def initialize_not_gpu_direct(shape: ArrayShape, dtype: np.dtype,
                                  tensor_format: TensorFormat, cudnn_dtype: int,
                                  tensor_type: TensorTypeEnum = TensorTypeEnum.TENSOR,
                                  desc: int | None = None, gpudirect: bool = False, cublas: bool = False) -> tuple[np.ndarray, "TensorGPU"]:
        x_cpu = np.zeros(shape, dtype)
        x_gpu = gpuarray.empty(shape, dtype)

        x_gpu = TensorGPU(x_gpu, tensor_format=tensor_format, cudnn_dtype=cudnn_dtype, tensor_type=tensor_type,
                          desc=desc, gpudirect=gpudirect, cublas=cublas)

        return (x_cpu, x_gpu)
    
    @staticmethod
    def initialize(shape: ArrayShape, dtype: np.dtype,
                   tensor_format: TensorFormat, cudnn_dtype: int,
                   tensor_type: TensorTypeEnum = TensorTypeEnum.TENSOR,
                   desc: int | None = None, gpudirect: bool = False, cublas: bool = False,
                   drv: "pycuda_driver" = None) -> tuple[np.ndarray, "TensorGPU"]:
        if drv is not None:
            return TensorGPU.initialize_gpu_direct(drv=drv, shape=shape,
                                                   dtype=dtype, tensor_format=tensor_format,
                                                   cudnn_dtype=cudnn_dtype, tensor_type=tensor_type,
                                                   desc=desc, gpudirect=gpudirect, cublas=cublas)
        else:
            return TensorGPU.initialize_not_gpu_direct(shape=shape, dtype=dtype, tensor_format=tensor_format,
                                                       cudnn_dtype=cudnn_dtype, tensor_type=tensor_type,
                                                       desc=desc, gpudirect=gpudirect, cublas=cublas)


    # ---

    def __init__(self, gpu_arr: "gpuarray.GPUArray", tensor_format: TensorFormat, cudnn_dtype: int,
                 tensor_type: TensorTypeEnum = TensorTypeEnum.TENSOR, desc: int | None = None,
                 gpudirect: bool = False, cublas: bool = False):

        self.tensor_format = TensorFormat(tensor_format.lower())
        self.cudnn_dtype = cudnn_dtype
        self.tensor_type = tensor_type
        self.gpudirect = gpudirect
        self.cublas = cublas
        self.cudnn_tensor_format = cudnn.cudnnTensorFormat['CUDNN_TENSOR_' + tensor_format.upper()]
        # The following atributes will be initalized in _initalize:
        self.ary: gpuarray.GPUArray = None
        self.size: int = -1
        self.desc: int = -1
        # ---
        self._initalize(gpu_arr, desc)
    # ---

    def copy(self):
        """ NumPy-like copy. """
        return copy.deepcopy(self)

    def __copy__(self):
        return TensorGPU(gpu_arr=self.ary,
                         tensor_format=self.tensor_format,
                         cudnn_dtype=self.cudnn_dtype,
                         tensor_type=self.tensor_type,
                         gpudirect=self.gpudirect,
                         cublas=self.cublas,
                         desc=self.desc)

    def __deepcopy__(self, memo: dict):
        obj = TensorGPU(gpu_arr=copy.deepcopy(self.ary, memo),
                        tensor_format=self.tensor_format,
                        cudnn_dtype=self.cudnn_dtype,
                        tensor_type=self.tensor_type,
                        gpudirect=self.gpudirect,
                        cublas=self.cublas,
                        desc=-1)
        memo[id(self)] = obj
        return obj

    def encode_shape(self, shape):
        return encode_shape(shape, self.tensor_format)

    def decode_shape(self, shape):
        return decode_shape(shape, self.tensor_format)

    def _set_shape(self, gpu_arr: "gpuarray.GPUArray") -> None:

        match len(gpu_arr.shape):
            case 1:
                match self.tensor_format:
                    case TensorFormat.NCHW:
                        self.shape = (1, *gpu_arr.shape, 1, 1)
                    case TensorFormat.NHWC:
                        self.shape = (1, 1, 1, *gpu_arr.shape)
                    case tensor_format:
                        raise NotImplementedError(f"Unsupported tensor format {tensor_format}!")
            case 2:
                match self.tensor_format:
                    case TensorFormat.NCHW:
                        self.shape = (*gpu_arr.shape, 1, 1)
                    case TensorFormat.NHWC:
                        self.shape = (gpu_arr.shape[0], 1, 1, gpu_arr.shape[1])
                    case tensor_format:
                        raise NotImplementedError(f"Unsupported tensor format {tensor_format}!")
            case 3:
                match self.tensor_format:
                    case TensorFormat.NCHW:
                        self.shape = (gpu_arr.shape[0], 1, gpu_arr.shape[1], gpu_arr.shape[2])
                    case TensorFormat.NHWC:
                        raise NotImplementedError("Shape padding not implemented for 3-dim shape on NHWC")
            case 4:
                self.shape = gpu_arr.shape
            case _:
                raise ValueError(f"The expected len shape are 1, 2 or 4. Shape received: {len(gpu_arr.shape)}.")
    # ---

    def _set_ptr(self, gpu_arr: "gpuarray.GPUArray") -> None:
        if self.gpudirect:
            self.ptr_intp = np.intp(self.ary.base.get_device_pointer())
            self.ptr = ctypes.c_void_p(int(self.ary.base.get_device_pointer()))
        else:
            self.ptr = ctypes.c_void_p(int(gpu_arr.gpudata))
    # ---

    def _set_desc(self, desc: int | None) -> None:
        if desc is not None:
            self.desc = desc
        else:
            match self.tensor_type:
                case self.TensorTypeEnum.TENSOR:
                    n, c, h, w = self.decode_shape(self.shape)
                    self.desc = cudnn.cudnnCreateTensorDescriptor()
                    cudnn.cudnnSetTensor4dDescriptor(self.desc, self.cudnn_tensor_format,
                                                     self.cudnn_dtype, n, c, h, w)
                case self.TensorTypeEnum.FILTER:
                    n, c, h, w = self.decode_shape(self.shape)
                    self.desc = cudnn.cudnnCreateFilterDescriptor()
                    cudnn.cudnnSetFilter4dDescriptor(self.desc, self.cudnn_dtype,
                                                     self.cudnn_tensor_format, n, c, h, w)
                case self.TensorTypeEnum.SEQ:
                    if len(self.shape) == 3:
                        self.shape = (self.shape[0], 1, self.shape[-2], self.shape[-1])
                    self.desc = cudnn.cudnnCreateSeqDataDescriptor()
                    dimA = np.array([0, 0, 0, 0], dtype=np.int32)
                    dimA[cudnn.cudnnSeqDataAxis["CUDNN_SEQDATA_BATCH_DIM"]] = self.shape[0]
                    dimA[cudnn.cudnnSeqDataAxis["CUDNN_SEQDATA_BEAM_DIM"]] = self.shape[1]
                    dimA[cudnn.cudnnSeqDataAxis["CUDNN_SEQDATA_TIME_DIM"]] = self.shape[2]
                    dimA[cudnn.cudnnSeqDataAxis["CUDNN_SEQDATA_VECT_DIM"]] = self.shape[3]
                    axes = np.array([0, 0, 0, 0], dtype=np.int32)
                    axes[0] = cudnn.cudnnSeqDataAxis["CUDNN_SEQDATA_BATCH_DIM"]
                    axes[1] = cudnn.cudnnSeqDataAxis["CUDNN_SEQDATA_BEAM_DIM"]
                    axes[2] = cudnn.cudnnSeqDataAxis["CUDNN_SEQDATA_TIME_DIM"]
                    axes[3] = cudnn.cudnnSeqDataAxis["CUDNN_SEQDATA_VECT_DIM"]
                    self.seq_length_array = np.full(shape=(self.shape[0] * self.shape[1]), fill_value=self.shape[-2], dtype=np.int32)
                    # print(self.shape, dimA, axes, len(seq_length_array))
                    cudnn.cudnnSetSeqDataDescriptor(self.desc, cudnn_dtype,
                                                    np.int32(4), dimA, axes,
                                                    np.int32(len(self.seq_length_array)), self.seq_length_array,
                                                    None)
                case self.TensorTypeEnum.OTHER:
                    pass  # do nothing.

                case tensor_type:
                    raise NotImplementedError(f"Tensor type not implemented! ({tensor_type})")
    # ---

    def _del_desc(self) -> None:
        match self.tensor_type:
            case self.TensorTypeEnum.TENSOR:
                cudnn.cudnnDestroyTensorDescriptor(self.desc)
            case self.TensorTypeEnum.FILTER:
                cudnn.cudnnDestroyFilterDescriptor(self.desc)
            case self.TensorTypeEnum.SEQ:
                pass
            case self.TensorTypeEnum.OTHER:
                cudnn.cudnnDestroySeqDataDescriptor(self.desc)
            case tensor_type:
                raise NotImplementedError(f"Tensor type not implemented! ({tensor_type})")
        self.desc = -1

    def _initalize(self, gpu_arr: "gpuarray.GPUArray", desc: int | None = None) -> None:
        self.ary = gpu_arr
        self._set_shape(gpu_arr)
        self.size = gpu_arr.size
        if self.size != 0:
            self._set_ptr(gpu_arr)
            self._set_desc(desc)
    # ---

    def reshape(self, shape: ArrayShape):
        self.ary = self.ary.reshape(shape, order="C")
        self.shape = shape
        return self
    # ---

    def free_gpu_arr(self) -> None:
        if self.ary is not None:
            self._del_desc()
            del self.ary
        self.size = -1
        self.desc = -1
    # ---

    def __del__(self) -> None:
        """Best effort finalizer"""
        try:
            self.free_gpu_arr()
        except:  # noqa: E722
            pass

    def set_ary(self, gpu_arr: "gpuarray.GPUArray", desc: int | None = None) -> None:
        self.free_gpu_arr()
        self._initalize(gpu_arr, desc)
    # ---

    def set_ary_from_ndarray(self, arr: np.ndarray, desc: int | None = None) -> None:
        self.free_gpu_arr()
        self._initalize(gpuarray.to_gpu(arr), desc)
    # ---

    def fill(self, scalar: int | float) -> None:
        self.ary.fill(scalar)
    # ---



from pathlib import Path
import tarfile
from typing import TYPE_CHECKING

import numpy as np
import csv

from itertools import chain
from pydtnn.datasets.dataset import Dataset
from pydtnn.utils import random

if TYPE_CHECKING:
    from pydtnn.model import Model


TRAIN_NSAMPLES = 86524
TEST_NSAMPLES = 25596
INPUT_SHAPE = (1, 1024, 1024)
OUTPUT_SHAPE = (15,)

CSV_DELIMETER = ','
CSV_LABELS_DELIMETER = '|'
CSV_IMAGES_FIELD = "Image Index"
CSV_LABELS_FIELD = "Finding Labels"
type Class = np.ndarray


def get_dict_file_labels(path: Path) -> dict[str, list[str]]:
    with open(file=path, mode="r") as file:
        reader = csv.DictReader(file, delimiter=CSV_DELIMETER)
        dict_file_labels = dict[str, list[str]]()
        for row in reader:
            image = row[CSV_IMAGES_FIELD]
            labels = row[CSV_LABELS_FIELD].split(CSV_LABELS_DELIMETER)
            dict_file_labels[image] = labels
    return dict_file_labels


class ChestXRay(Dataset):
    """
    ChestXRay Dataset

    The NIH Chest X-ray dataset consists of 100,000 de-identified
    images of chest x-rays. The images are in PNG format.

    Source (SHA1): https://nihcc.app.box.com/v/ChestXray-NIHCC
    48a9f849a8f100a0f1721b33bdbd209767656111 Data_Entry_2017_v2020.csv
    41b85e218abec560a2f5999acbcf333b0f2fa495 test_list.txt
    e3e1b677c01d28481777f3d84e10fcdaac05694c train_val_list.txt
    fef95a7a789bcb0013fbf966cb92c4d92c90becd images_01.tar.gz https://nihcc.box.com/shared/static/vfk49d74nhbxq3nqjg0900w5nvkorp5c.gz
    23d2f2cd62d0271b16869abcdd8a00a1fd2492b5 images_02.tar.gz https://nihcc.box.com/shared/static/i28rlmbvmfjbl8p2n3ril0pptcmcu9d1.gz
    69935ac7886c18246446899cb2b75443195847c4 images_03.tar.gz https://nihcc.box.com/shared/static/f1t00wrtdk94satdfb9olcolqx20z2jp.gz
    fb86b3adaad0e9ff1405154cf6521e180063af10 images_04.tar.gz https://nihcc.box.com/shared/static/0aowwzs5lhjrceb3qp67ahp0rd1l1etg.gz
    baa8155f0285edb4a07e717f79682713416eb205 images_05.tar.gz https://nihcc.box.com/shared/static/v5e3goj22zr6h8tzualxfsqlqaygfbsn.gz
    3a4252d82143757600885121bb57b0ef4e482532 images_06.tar.gz https://nihcc.box.com/shared/static/asi7ikud9jwnkrnkj99jnpfkjdes7l6l.gz
    cd3cd855acb4e12ca11608be6aae99414d4bc22b images_07.tar.gz https://nihcc.box.com/shared/static/jn1b4mw4n6lnh74ovmcjb8y48h8xj07n.gz
    d8891e0079e88fc04dab45253b86a2214ff499b6 images_08.tar.gz https://nihcc.box.com/shared/static/tvpxmn7qyrgl0w8wfh9kqfjskv6nmm1j.gz
    84661300d777e07be9ae7d5f37fb82721202f0bc images_09.tar.gz https://nihcc.box.com/shared/static/upyy3ml7qdumlgk2rfcvlb9k6gvqq2pj.gz
    30216f59778f259db91d77bcd3d0495c8fce88ef images_10.tar.gz https://nihcc.box.com/shared/static/l6nilvfa9cg3s28tqv1qc1olm3gnz54p.gz
    97985118ba36f18c27d62371d28c1698478cecfa images_11.tar.gz https://nihcc.box.com/shared/static/hhq8fkdgvcari67vfhs7ppg2w6ni4jze.gz
    cb2865369f434a9deea11e2d5222b8472890681b images_12.tar.gz https://nihcc.box.com/shared/static/ioqwiy20ihqwyr8pf4c24eazhh281pbu.gz

    Normalize (z-score):
    offset: -0.509
    scale:  +4.002
    """

    def __init__(self, model: "Model", force_test_as_validation=False, debug=False):
        super().__init__(model, TRAIN_NSAMPLES, TEST_NSAMPLES, INPUT_SHAPE, OUTPUT_SHAPE, force_test_as_validation=force_test_as_validation, debug=debug)

    def _get_labels(self, image_file: str) -> Class:
        labels = self._dict_images_labels[image_file]
        mask = np.zeros(self.output_shape, dtype=np.uint8)
        for label in labels:
            mask[self.labels2classes[label]] = 1
        return mask

    def _init_actual_data(self):
        self._xy_filenames: list[list[tuple[str, Class]]] = [[("", np.empty((0,)))] for _ in Dataset.Part]
        self.files = Path(self.model.dataset_path)
        csv = self.files.joinpath("Data_Entry_2017_v2020.csv")
        train = self.files.joinpath("train_val_list.txt").read_text().splitlines()
        test = self.files.joinpath("test_list.txt").read_text().splitlines()
        self._dict_images_labels = get_dict_file_labels(csv)

        output_shape = len(set(chain.from_iterable(self._dict_images_labels.values())))

        if output_shape != OUTPUT_SHAPE[0]:
            raise ValueError(f"Mismatch class shape (got: {output_shape}, expect: {OUTPUT_SHAPE[0]})")

        if len(train) != TRAIN_NSAMPLES:
            raise ValueError(f"Mismatch train samples (got: {len(train)}, expect: {TRAIN_NSAMPLES})")

        if len(test) != TEST_NSAMPLES:
            raise ValueError(f"Mismatch train samples (got: {len(test)}, expect: {TEST_NSAMPLES})")

        # Getting the labels and equivalence class - label
        labels = sorted(list({elem for list_elems in self._dict_images_labels.values() for elem in list_elems}))
        self.labels2classes = {labels[_class]: _class for _class in range(len(labels))}

        # Get image map
        self._src_filename = dict[str, str]()
        for gz in self.files.glob("*.gz"):
            with self._gzip_open(str(gz)) as g:
                with tarfile.TarFile(fileobj=g) as t:
                    for member in t.getmembers():
                        if not member.isfile():
                            continue
                        self._src_filename[member.name] = str(gz)

        # Format: 'datasets/chestxray/images_06.tar.gz', 'images/00011575_000.png'
        self._xy_filenames[Dataset.Part.TRAIN] = [(f"images/{name}", self._get_labels(name)) for name in train]
        self._xy_filenames[Dataset.Part.TEST] = [(f"images/{name}", self._get_labels(name)) for name in test]
        self._xy_filenames[Dataset.Part.VAL] = self._xy_filenames[Dataset.Part.TEST] if self.model.test_as_validation else self._xy_filenames[Dataset.Part.TRAIN]

    def _actual_data_generator(self, part):
        offset = self._local_offset[part]
        nsamples = self._local_nsamples[part]
        xy_filenames = self._xy_filenames[part]

        if part is Dataset.Part.TRAIN and self.model.augment_shuffle:
            random.shuffle(xy_filenames)  # type: ignore (numpy shuffle's typing wasn't well defined.)

        xy_filenames = xy_filenames[offset:offset + nsamples]

        for path, y in xy_filenames:
            src_path = self._src_filename[path]
            with self._gzip_open(src_path) as g, tarfile.TarFile(fileobj=g) as t, t.extractfile(path) as fp:
                x = self._load_gray_image(fp)

            # Add N dimension
            x = x[None, ...]
            y = y[None, ...]

            # Set tensor format
            x = self.model.encode_tensor(x)

            # Set dtype and order
            x = x.astype(dtype=self.model.dtype, order="C")
            y = y.astype(dtype=self.model.dtype, order="C")

            # Inplace normalization
            x /= 255.0

            yield x, y



from typing import TYPE_CHECKING

import numpy as np

from pydtnn.optimizers.optimizer import Optimizer
from pydtnn.utils.constants import Array

if TYPE_CHECKING:
    from pydtnn.model import Model


class Adam[T: Array](Optimizer[T]):
    """
    Adam optimizer
    """

    def __init__(self, learning_rate: float = 1e-2, beta1: float = 0.99, beta2: float = 0.999,
                 epsilon: float = 1e-7, decay: float = 0.0, dtype: np.dtype = np.dtype(np.float32)):
        super().__init__(learning_rate=learning_rate, dtype=dtype)
        self.beta1: float = beta1
        self.beta2: float = beta2
        self.epsilon: float = epsilon
        self.decay: float = decay

    @classmethod
    def from_model(cls, model: "Model") -> "Adam":
        return Adam(learning_rate=model.learning_rate,
                    beta1=model.optimizer_beta1,
                    beta2=model.optimizer_beta2,
                    epsilon=model.optimizer_epsilon,
                    decay=model.optimizer_decay,
                    dtype=model.dtype)



import numpy as np

from pydtnn.layers.layer import Layer, LayerError
from pydtnn.utils.constants import Array


class AdaptiveAveragePool2D[T: Array](Layer):

    # This layer will calculate the pool shape and the stride from the output shape (passed as parameter) and the previous layer shape.

    # output_shape:
    #  -> None: if the output shape is equal to the input
    #  -> int: if all the output shape's dimensions share values
    #  -> Tuple[int, int]: if it is necessary or it is preferred to define each output dimension individually

    def __init__(self, output_shape: int | tuple[int, int] | None = None):
        super().__init__()
        self.output_shape = output_shape

        # This value will change in initialize:
        self.pooling_not_needed: bool = None  # type: ignore
    # ---  END __init__ --- #

    def initialize(self, prev_shape: tuple[int, int], x: T | None) -> None:
        super().initialize(prev_shape, x)

        self.ci, self.hi, self.wi = self.model.decode_shape(prev_shape)

        if self.output_shape is None:
            self.ho, self.wo = self.hi, self.wi
        else:
            self.ho, self.wo = (self.output_shape, self.output_shape) if isinstance(self.output_shape, int) else self.output_shape

        if not (self.ho > 0 and self.wo > 0):
            raise LayerError(f"The output height and width should be grater than 0. height: {self.ho} width: {self.wo}")
        self.co = self.ci

        # If the output and the input shapes are the same, there is no need of pooling.
        self.pooling_not_needed = (self.hi == self.ho) and (self.wi == self.wo)

        self.shape = self.model.encode_shape((self.co, self.ho, self.wo))
        self.n = np.prod(self.shape)
    # - END initialize - #



from typing import TYPE_CHECKING

from pydtnn.schedulers.scheduler import Scheduler

from numpy import ndarray

if TYPE_CHECKING:
    from pydtnn.model import Model


class WarmUp(Scheduler):
    """
    WarmUp
    """

    def __init__(self, warmup_epochs=5, base_lr=1e-4, init_lr=1e-3, verbose=False):
        super().__init__(verbose)
        self.warmup_epochs = warmup_epochs
        self.base_lr = base_lr
        self.init_lr = init_lr
        self.epoch_count: int = 0

    def on_epoch_end(self, train_loss: ndarray[float], val_loss: ndarray[float]) -> None:
        if self.epoch_count < self.warmup_epochs:
            self.model.optimizer.learning_rate = self.base_lr + ((self.epoch_count + 1) / self.warmup_epochs) * (self.init_lr - self.base_lr)
            self.epoch_count += 1
            self.log(f"Setting learning rate to {self.model.optimizer.learning_rate:.8f}.")

    @classmethod
    def from_model(cls, model: "Model") -> "WarmUp":
        return WarmUp(model.warm_up_epochs,
                      model.learning_rate / model.nprocs,
                      model.learning_rate)



from collections.abc import Sequence

from pydtnn.activations.relu import Relu
from pydtnn.activations.softmax import Softmax
from pydtnn.layers.batch_normalization import BatchNormalization
from pydtnn.layers.conv_2d import Conv2D
from pydtnn.layers.dropout import Dropout
from pydtnn.layers.fc import FC
from pydtnn.layers.flatten import Flatten
from pydtnn.layers.input import Input
from pydtnn.layer_base import LayerBase
from pydtnn.utils.initializers import he_uniform
from pydtnn.layers.max_pool_2d import MaxPool2D


def vgg3dobn(input_shape: Sequence[int], output_shape: Sequence[int]) -> Sequence[LayerBase]:
    model = list[LayerBase]()
    _ = model.append

    _(Input(shape=input_shape))
    for n_filt, do_rate in zip([32, 64, 128], [0.2, 0.3, 0.4]):
        for i in range(2):
            _(Conv2D(nfilters=n_filt, filter_shape=(3, 3), padding=1, weights_initializer=he_uniform))
            _(Relu())
            _(BatchNormalization())
        _(MaxPool2D(pool_shape=(2, 2), stride=2))
        _(Dropout(rate=do_rate))
    _(Flatten())
    _(FC(shape=(512,), weights_initializer=he_uniform))
    _(Relu())
    _(BatchNormalization())
    _(Dropout(rate=0.5))
    _(FC(shape=output_shape, weights_initializer=he_uniform))
    _(Softmax())

    return model



import numpy as np
from pydtnn.model import Model
from pydtnn.layers.input import Input
from pydtnn.layers.layer import Layer
from pydtnn.activations.leaky_relu import LeakyRelu
from pydtnn.activations.relu6 import Relu6
from copy import deepcopy

from pydtnn.layers.conv_2d import Conv2D
from pydtnn.utils import random

from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
try:
    import pycuda.gpuarray as gpuarray
    from pydtnn.libs import libcudnn as cudnn
except BaseException:
    pass

# Constants #
TENSOR_FORMAT = "NCHW"  # "NCHW" # "NHWC" # "NCHW"
N, C, H, W = 2, 2, 3, 3
SHAPE = (C, H, W) if TENSOR_FORMAT == "NCHW" else (H, W, C)
CONV_IN_CHANNELS = C
CONV_OUT_CHANNELS = 2  # = PyTorch's Number filters
CONV_KERNEL_SIZE = (1, 1)
SEED = 1234
DTYPE = np.float32

KWARGS = {
    "model_name": None,
    "evaluate_only": True,
    "parallel": "data",
    "tensor_format": TENSOR_FORMAT,
    "enable_gpu": False,  # True,
    "omm": None,
    "dtype": DTYPE,
    "tracing": False,
    "tracer_output": "",
    "batch_size": min(64, N),
    "optimizer_name": "adam",
}

# End Constants #

random.seed(SEED)


def main():
    model_I2C = Model(**KWARGS)
    model_DEPTH = Model(**KWARGS)
    model_POINT = Model(**KWARGS)
    model_RELU = Model(**KWARGS)

    shape = (N, *SHAPE)
    dataset = np.arange(np.prod(shape), dtype=DTYPE).reshape(shape)
    dataset *= -1
    dataset *= dataset % 2
    print(f"{dataset=}")
    print(f"{dataset.shape}")

    use_bias = True

    models = [
        ("=============\n==== I2C ====\n=============", model_I2C),
        ("=============\n= POINTWISE =\n=============", model_POINT),
        ("=============\n= DEPTHWISE =\n=============", model_DEPTH),
        ("=============\n= LEAKY RELU =\n=============", model_RELU)
    ]

    for _, model in models:
        model: Model
        model.dataset = dataset

    model_RELU.add(Input(SHAPE, is_shape_in_format=True))
    model_RELU.add(LeakyRelu(negative_slope=-32))
    model_RELU.add(Relu6())
    model_RELU._initialize()

    model_DEPTH.add(Input(SHAPE, is_shape_in_format=True))
    model_DEPTH.add(Conv2D(nfilters=CONV_OUT_CHANNELS, filter_shape=CONV_KERNEL_SIZE, grouping=Conv2D.Grouping.DEPTHWISE, use_bias=use_bias))
    model_DEPTH._initialize()

    model_POINT.add(Input(SHAPE, is_shape_in_format=True))
    model_POINT.add(Conv2D(nfilters=CONV_OUT_CHANNELS, filter_shape=CONV_KERNEL_SIZE, grouping=Conv2D.Grouping.POINTWISE, use_bias=use_bias))
    model_POINT._initialize()

    model_I2C.add(Input(SHAPE, is_shape_in_format=True))
    model_I2C.add(Conv2D(nfilters=CONV_OUT_CHANNELS, filter_shape=CONV_KERNEL_SIZE, use_bias=use_bias))
    model_I2C._initialize()

    for name, model in models:
        print(f"{name}")

        model.mode = Model.Mode.TRAIN
        # model.show()

        x = deepcopy(dataset)
        if KWARGS["enable_gpu"]:
            _dataset = TensorGPU(
                gpu_arr=gpuarray.empty(shape=dataset.shape, dtype=KWARGS["dtype"]),
                tensor_format=model.tensor_format, cudnn_dtype=model.cudnn_dtype)
            _dataset.ary.set(dataset)
            x = _dataset

        num_layers = len(model.layers)
        print("Forward")
        for i in range(num_layers):
            layer: Layer = model.layers[i]
            print(f"{layer=}")
            x: np.ndarray | TensorGPU = layer.forward(x)
            print(f"{x.shape=}")
        print("\n----------")

        dy = x
        print("Backward")
        for i in range(num_layers - 1, 0, -1):
            layer: Layer = model.layers[i]
            dy: np.ndarray | TensorGPU = layer.backward(dy)
            print(f"{dy.shape=}")
        print("\n=========\n")

        for i in range(num_layers - 1, 0, -1):
            layer: Layer = model.layers[i]
            layer.update_weights(model.optimizer)


if __name__ == "__main__":
    main()



from pydtnn.activations.tanh import Tanh
from pydtnn.backends.gpu.activations.activation import ActivationGPU
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.utils.constants import ArrayShape

from pydtnn.libs import libcudnn as cudnn
import pycuda.gpuarray as gpuarray  # type: ignore


class TanhGPU(Tanh[TensorGPU], ActivationGPU):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.act_desc = None

    def initialize(self, prev_shape: ArrayShape, x: TensorGPU) -> None:
        super().initialize(prev_shape, x)

        self.act_desc = cudnn.cudnnCreateActivationDescriptor()

        mode = cudnn.cudnnActivationMode['CUDNN_ACTIVATION_TANH']
        nan = cudnn.cudnnNanPropagation['CUDNN_NOT_PROPAGATE_NAN']
        cudnn.cudnnSetActivationDescriptor(self.act_desc, mode, nan, 0.0)

        # Activations y
        y_gpu = gpuarray.empty(x.ary.shape, self.model.dtype)
        self.y = TensorGPU(y_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        # Derivative dx
        dx_gpu = gpuarray.empty(x.ary.shape, self.model.dtype)
        self.dx = TensorGPU(dx_gpu, self.model.tensor_format, self.model.cudnn_dtype)

    def forward(self, x: TensorGPU) -> TensorGPU:
        alpha, beta = 1.0, 0.0
        cudnn.cudnnActivationForward(self.model.cudnn_handle, self.act_desc, alpha,
                                     x.desc, x.ptr, beta,
                                     self.y.desc, self.y.ptr)
        return self.y

    def backward(self, dy: TensorGPU) -> TensorGPU:
        alpha, beta = 1.0, 0.0
        cudnn.cudnnActivationBackward(self.model.cudnn_handle, self.act_desc, alpha,
                                      self.y.desc, self.y.ptr,
                                      dy.desc, dy.ptr,
                                      self.x.desc, self.x.ptr, beta,
                                      self.dx.desc, self.dx.ptr)
        return self.dx



import atexit
from collections import defaultdict
from timeit import default_timer as timer

from pydtnn.tracers.tracer import Tracer

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pympi.MPI import Comm as MPI_COMM
else:
    from types import ModuleType
    MPI_COMM = ModuleType


class SimpleTracer(Tracer):
    """
    SimpleTracer
    """

    def __init__(self, tracing: bool, output_filename: str, comm: MPI_COMM | None):
        super().__init__(tracing)
        self.output_filename = output_filename
        self.rank = 0
        if comm is not None:
            self.rank = comm.Get_rank()
        self.events = defaultdict(lambda: defaultdict(lambda: [0, []]))
        self.pending_events = []

    def enable_tracing(self):
        super().enable_tracing()
        # If tracing is enabled at least once, register self.write_output to be executed at exit
        atexit.register(self._write_output)

    def _emit_event(self, evt_type_val: int, evt_val: int, stream=None):
        """This method will be called only if tracing is enabled"""
        if evt_val != 0:
            self.pending_events.append((evt_type_val, evt_val, timer()))
        else:
            toc = timer()
            if len(self.pending_events) == 0:
                raise RuntimeError("Received an 'End' event but there are no pending events!")
            if self.pending_events[-1][0] != evt_type_val:
                raise RuntimeError("Received an 'End' event for a different event type than expected!")
            _evt_type_val, _evt_val, tic = self.pending_events.pop()
            self.events[_evt_type_val][_evt_val][0] += 1
            self.events[_evt_type_val][_evt_val][1].append(toc - tic)

    def _emit_nevent(self, evt_type_val_list: list[int], evt_val_list: list[int], stream=None):
        """This method will be called only if tracing is enabled"""
        zipped_list = list(zip(evt_type_val_list, evt_val_list))
        if evt_val_list[0] == 0:
            zipped_list = reversed(zipped_list)
        for evt_type_val, evt_val in zipped_list:
            self.emit_event(evt_type_val, evt_val)

    def _output_header(self) -> str:
        return "Event type;Event value;Event name;Calls;Total time;Median of times"

    def _output_row(self, event_type_value: int, event_value: int) -> str:
        event_type = self.event_types[event_type_value]
        event_type_name = event_type.name
        _calls, _times = self.events[event_type_value][event_value]
        _times.sort()
        total_time = sum(_times)
        mean_of_times = _times[len(_times) // 2]
        return f"{event_type_name};{event_value};{event_type[event_value]};{_calls};{total_time};{mean_of_times}"

    def _write_output(self):
        """This method will be called at exit only if tracing has been enabled at any time"""
        if self.rank == 0:
            if len(self.pending_events):
                print("Warning: finishing simple tracer while there are pending events to be marked as finished.")
            print(f"Writing simple tracer output to '{self.output_filename}'...")
            with open(self.output_filename, 'w') as f:
                f.write(self._output_header() + "\n")
                for event_type_value, events in self.events.items():
                    for event_value in events.keys():
                        f.write(self._output_row(event_type_value, event_value) + "\n")



# https://github.com/storypku/cuda-support-for-bazel/blob/9a9c90c7d73fdafb3fbc8713232405cae4ae66d8/examples/cudnn-samples/multiHeadAttention/multiHeadAttention.cpp
import numpy as np

from pydtnn.libs import libcudnn as cudnn
import pycuda.gpuarray as gpuarray
import pycuda

from pydtnn.backends.gpu.layers.layer import LayerGPU
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.layers.multi_head_attention import MultiHeadAttention
from pydtnn.utils import initializers
from pydtnn.backends.gpu.utils.memory_allocation import checkConvolutionMemory, getConvolutionWorkspaceSize, getConvolutionWorkspacePtr


class MultiHeadAttentionGPU(MultiHeadAttention[TensorGPU], LayerGPU):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.beam = 1  # Number of hypothesis we keep

        self.grad_vars = {"weights": "dw"}

        # The next attributes will be initialized later
        self.y = self.dx = None

    def initialize(self, prev_shape, x):
        super().initialize(prev_shape, x)
        self.query, self.key, self.value, self.mask = x
        self.shape = prev_shape

        self.batch = self.query.shape[0]
        self.seq = self.query.shape[-2]
        self.embedl = self.query.shape[-1]

        # Weights Initializers
        weights_shape = (self.embedl, self.heads*self.d_k)
        biases_shape = (1, self.heads*self.d_k)
        o_weights_shape = (self.heads*self.d_k, self.embedl)
        o_biases_shape = (1, self.embedl)

        self.q_weights_cpu = self.weights_initializer(weights_shape, self.model.dtype)
        self.k_weights_cpu = self.weights_initializer(weights_shape, self.model.dtype)
        self.v_weights_cpu = self.weights_initializer(weights_shape, self.model.dtype)
        self.q_biases_cpu = self.biases_initializer(biases_shape, self.model.dtype)
        self.k_biases_cpu = self.biases_initializer(biases_shape, self.model.dtype)
        self.v_biases_cpu = self.biases_initializer(biases_shape, self.model.dtype)
        self.o_weights_cpu = self.weights_initializer(o_weights_shape, self.model.dtype)
        self.o_biases_cpu = self.biases_initializer(o_biases_shape, self.model.dtype)
        self.nparams = 0
        _weights = [self.q_weights_cpu, self.k_weights_cpu, self.v_weights_cpu, self.o_weights_cpu,
                    self.q_biases_cpu, self.k_biases_cpu, self.v_biases_cpu, self.o_biases_cpu]
        for w in _weights:
            self.nparams += w.size

        # Dropout Descriptor
        self.states_size = cudnn.cudnnDropoutGetStatesSize(self.model.cudnn_handle)
        states_gpu = gpuarray.empty((self.states_size.value,), self.model.dtype)
        self.states = TensorGPU(states_gpu, self.model.tensor_fmt, self.model.cudnn_dtype, TensorGPU.TensorTypeEnum.OTHER)
        self.drop_desc = cudnn.cudnnCreateDropoutDescriptor()
        cudnn.cudnnSetDropoutDescriptor(self.drop_desc, self.model.cudnn_handle, self.dropout_rate,
                                        self.states.ptr, self.states_size.value, seed=0)

        # Attention Descriptor
        self.attn_mode = cudnn.cudnnAttnMode["CUDNN_ATTN_QUERYMAP_ONE_TO_ONE"]
        self.attn_mode += cudnn.cudnnAttnMode["CUDNN_ATTN_ENABLE_PROJ_BIASES"]
        self.add_grad = cudnn.cudnnWgradMode["CUDNN_WGRAD_MODE_SET"]
        self.attn_desc = cudnn.cudnnCreateAttnDescriptor()
        cudnn.cudnnSetAttnDescriptor(self.attn_desc, self.attn_mode, self.heads, 1.0, self.model.cudnn_dtype, self.model.cudnn_dtype, cudnn.cudnnMathType['CUDNN_DEFAULT_MATH'],
                                     None, None, self.embedl, self.embedl, self.embedl, self.d_k, self.d_k, self.d_k, self.embedl,
                                     self.seq, self.seq, self.model.batch_size, self.beam)

        # GPU Memory Allocation
        self.weights_size, self.workspace_size, self.reserve_backward_size = cudnn.cudnnGetMultiHeadAttnBuffers(self.model.cudnn_handle, self.attn_desc)
        checkConvolutionMemory(self.workspace_size)
        weights_size = self.weights_size.value // np.dtype(self.model.dtype).itemsize
        reserve_backward_size = self.reserve_backward_size.value // np.dtype(self.model.dtype).itemsize + 1
        self.weights = gpuarray.empty((weights_size,), self.model.dtype)
        self.weights = TensorGPU(self.weights, self.model.tensor_fmt, self.model.cudnn_dtype, TensorGPU.TensorTypeEnum.OTHER)
        self.dw = gpuarray.empty((weights_size,), self.model.dtype)
        self.dw = TensorGPU(self.dw, self.model.tensor_fmt, self.model.cudnn_dtype, TensorGPU.TensorTypeEnum.OTHER)
        self.reserve_backward = gpuarray.empty((reserve_backward_size,), self.model.dtype)
        self.reserve_backward = TensorGPU(self.reserve_backward, self.model.tensor_fmt, self.model.cudnn_dtype, TensorGPU.TensorTypeEnum.OTHER)

        # Weights to GPU
        # self.copy_weights()
        _weights_types = ["CUDNN_MH_ATTN_Q_WEIGHTS", "CUDNN_MH_ATTN_K_WEIGHTS", "CUDNN_MH_ATTN_V_WEIGHTS", "CUDNN_MH_ATTN_O_WEIGHTS",
                          "CUDNN_MH_ATTN_Q_BIASES", "CUDNN_MH_ATTN_K_BIASES", "CUDNN_MH_ATTN_V_BIASES", "CUDNN_MH_ATTN_O_BIASES"]
        _weights = [self.q_weights_cpu, self.k_weights_cpu, self.v_weights_cpu, self.o_weights_cpu,
                    self.q_biases_cpu, self.k_biases_cpu, self.v_biases_cpu, self.o_biases_cpu]
        pycuda.driver.memcpy_htod(self.weights.ptr.value, np.concatenate([w.flatten() for w in _weights]))

        # Memory Allocation for Outputs
        self.y = gpuarray.empty((self.model.batch_size, self.beam, self.seq, self.embedl), self.model.dtype)
        self.y = TensorGPU(self.y, self.model.tensor_fmt, self.model.cudnn_dtype, TensorGPU.TensorTypeEnum.SEQ)
        self.dquery = gpuarray.empty((self.model.batch_size, self.beam,  self.seq, self.embedl), self.model.dtype)
        self.dquery = TensorGPU(self.dquery, self.model.tensor_fmt, self.model.cudnn_dtype, TensorGPU.TensorTypeEnum.SEQ)
        self.dkey = gpuarray.empty((self.model.batch_size,  self.beam, self.seq, self.embedl), self.model.dtype)
        self.dkey = TensorGPU(self.dkey, self.model.tensor_fmt, self.model.cudnn_dtype, TensorGPU.TensorTypeEnum.SEQ)
        self.dvalue = gpuarray.empty((self.model.batch_size,  self.beam, self.seq, self.embedl), self.model.dtype)
        self.dvalue = TensorGPU(self.dvalue, self.model.tensor_fmt, self.model.cudnn_dtype, TensorGPU.TensorTypeEnum.SEQ)

        self.current_index = -1  # Training
        self.low_window_index = np.full(shape=(self.batch, self.beam, self.seq), fill_value=0, dtype=np.int32)
        self.high_window_index = np.full(shape=(self.batch, self.beam, self.seq), fill_value=self.seq, dtype=np.int32)
        # self.dev_seq_lengths_QO = np.full(shape=(self.batch*self.beam), fill_value=self.seq, dtype=np.int32)
        self.dev_seq_lengths_QO = np.copy(self.y.seq_length_array)
        self.dev_seq_lengths_QO = gpuarray.to_gpu(self.dev_seq_lengths_QO)
        # self.dev_seq_lengths_KV = np.full(shape=(self.batch*self.beam), fill_value=self.seq, dtype=np.int32)
        self.dev_seq_lengths_KV = np.copy(self.dkey.seq_length_array)
        self.dev_seq_lengths_KV = gpuarray.to_gpu(self.dev_seq_lengths_KV)

    def copy_weights(self):
        _weights_types = ["CUDNN_MH_ATTN_Q_WEIGHTS", "CUDNN_MH_ATTN_K_WEIGHTS", "CUDNN_MH_ATTN_V_WEIGHTS", "CUDNN_MH_ATTN_O_WEIGHTS",
                          "CUDNN_MH_ATTN_Q_BIASES", "CUDNN_MH_ATTN_K_BIASES", "CUDNN_MH_ATTN_V_BIASES", "CUDNN_MH_ATTN_O_BIASES"]
        _weights = [self.q_weights_cpu, self.k_weights_cpu, self.v_weights_cpu, self.o_weights_cpu,
                    self.q_biases_cpu, self.k_biases_cpu, self.v_biases_cpu, self.o_biases_cpu]

        pycuda.driver.memcpy_htod(self.weights.ptr.value, np.concatenate([w.flatten() for w in _weights]))
        return

        for i in range(len(_weights)):
            wDesc, dest = cudnn.cudnnGetMultiHeadAttnWeights(self.model.cudnn_handle, self.attn_desc,
                                                             cudnn.cudnnMultiHeadAttnWeightKind[_weights_types[i]], self.weights_size.value, self.weights.ptr)
            print(_weights_types[i], dest)
            # Check wDesc order matches
            if dest is not None:
                pycuda.driver.memcpy_htod(dest[0], _weights[i])

    def forward(self, query, key, value, mask=None, residuals=None):
        if True:  # self.model.mode == Model.Mode.TRAIN:
            self.query, self.key, self.value = query, key, value
            # return self.query
            cudnn.cudnnMultiHeadAttnForward(self.model.cudnn_handle, self.attn_desc,
                                            self.current_index, self.low_window_index, self.high_window_index,
                                            self.dev_seq_lengths_QO.ptr, self.dev_seq_lengths_KV.ptr,
                                            self.dquery.desc, query.ptr,
                                            residuals.ptr,
                                            self.dkey.desc, key.ptr,
                                            self.dvalue.desc, value.ptr,
                                            self.y.desc, self.y.ptr,
                                            self.weights_size.value, self.weights.ptr,
                                            getConvolutionWorkspaceSize(), getConvolutionWorkspacePtr(),
                                            self.reserve_backward_size.value, self.reserve_backward.ptr)
        else:
            cudnn.cudnnMultiHeadAttnForward(self.model.cudnn_handle, self.attn_desc,
                                            self.current_index, self.low_window_index, self.high_window_index,
                                            self.dev_seq_lengths_QO.ptr, self.dev_seq_lengths_KV.ptr,
                                            self.dquery.desc, query.ptr,
                                            residuals.ptr,
                                            self.dkey.desc, key.ptr,
                                            self.dvalue.desc, value.ptr,
                                            self.y.desc, self.y.ptr,
                                            self.weights_size.value, self.weights.ptr,
                                            getConvolutionWorkspaceSize(), getConvolutionWorkspacePtr(),
                                            0, None)
        return self.y

    def backward(self, dy):
        cudnn.cudnnMultiHeadAttnBackwardData(self.model.cudnn_handle, self.attn_desc,
                                             self.low_window_index, self.high_window_index,
                                             self.dev_seq_lengths_QO.ptr, self.dev_seq_lengths_KV.ptr,
                                             self.y.desc, dy.ptr,
                                             self.dquery.desc, self.dquery.ptr, self.query.ptr,
                                             self.dkey.desc, self.dkey.ptr, self.key.ptr,
                                             self.dvalue.desc, self.dvalue.ptr, self.value.ptr,
                                             self.weights_size.value, self.weights.ptr,
                                             getConvolutionWorkspaceSize(), getConvolutionWorkspacePtr(),
                                             self.reserve_backward_size.value, self.reserve_backward.ptr)

        cudnn.cudnnMultiHeadAttnBackwardWeights(self.model.cudnn_handle, self.attn_desc, self.add_grad,
                                                self.dquery.desc, self.query.ptr,
                                                self.dkey.desc, self.key.ptr,
                                                self.dvalue.desc, self.value.ptr,
                                                self.y.desc, dy.ptr,
                                                self.weights_size.value, self.weights.ptr, self.dw.ptr,
                                                getConvolutionWorkspaceSize(), getConvolutionWorkspacePtr(),
                                                self.reserve_backward_size.value, self.reserve_backward.ptr)

        return self.dquery, self.dkey, self.dvalue



"""
https://github.com/google-research/bert
https://github.com/google-research/bert/blob/master/modeling.py
https://www.kaggle.com/tanulsingh077/deep-learning-for-nlp-zero-to-transformers-bert
https://huggingface.co/datasets/wikipedia
https://huggingface.co/albert-base-v2

d_k = embeld / heads

Tensorflow parameters names:
    hidden_size -> embedl
    intermediate_size -> d_ff
    num_hidden_layers -> n_encoders
    The rest of the parameters should be obvious
"""

from collections.abc import Sequence
from pydtnn.layer_base import LayerBase
from pydtnn.layers.encoder import Encoder
from pydtnn.layers.input import Input


def bert(input_shape: Sequence[int], output_shape: Sequence[int]) -> Sequence[LayerBase]:
    """Bert-Medium"""
    model = list[LayerBase]()
    _ = model.append

    n_encoders = 8
    max_seq = 512
    embedl = 512

    _(Input(shape=(1, max_seq, embedl)))
    for i in range(n_encoders):
        _(Encoder(embedl=embedl, d_k=64, heads=8, d_ff=2048, dropout_rate=0.1))

    return model

# def create_bert(model):
#     n_encoders = 3
#     _ = model.add
#     _(Input(shape=(1,75,300)))
#     for i in range(n_encoders):
#         _(Encoder(embedl=300, d_k=30, heads=10, d_ff=1200, dropout_rate=0.1))



"""
PyDTNN convDirect module
"""

import ctypes
import sys
import weakref

import numpy as np

from pydtnn.utils.tensor import TensorFormat, encode_shape, decode_shape
from pydtnn.utils import load_library
from pydtnn.backends.cpu.utils.im2row_nhwc_cython import im2row_nhwc_cython

try:
    load_library("convDirect")
    is_conv_direct_available = True
except Exception:
    is_conv_direct_available = False


class ConvDirect:
    """
    Exposes the libconvDirect functions following the PyDTNN conventions.

    Methods
    -------
    conv_direct(weights, x, out, vpadding, hpadding,
                vstride, hstride, vdilation, hdilation)
        calls the appropriate convDirect function from libconvDirect.so to perform
        the selected convDirect method.

    Examples
    --------
    See __usage_example__() method for an example of use. This example can be
    run with: 'python conv_direct.py'

    Tests
    -----
    To perform the tests, run the following command from the current directory:
        python -m unittest tests.convWinogradTestcase

    (see tests/winograd.py for more instructions on testing)
    """

    lib_cd = None  # will link to the libconvDirect.so library

    def _set_methods(self, method_name):
        return

    def __init__(self, method_name, dtype: np.dtype = np.dtype(np.float32), tensor_format=TensorFormat.NHWC,
                 debug=False, parent_layer=None):
        """
        Loads the libconvDirect.so library.

        Parameters
        ----------
        method : string
            Name of the method to be used
        dtype : data type
            The element data type being used on all the matrices.
        tensor_format : enumerate
            The tensor format to be used
        debug : boolean
            Whether to print debug information or not.
        parent_layer: layer
            The layer that is using it.
        """

        # self properties from parameters
        self.dtype = dtype
        self.tensor_format = tensor_format
        self.debug = debug

        # Parent layer
        if parent_layer is not None:
            self.get_parent_layer = weakref.ref(parent_layer)
            self.evaluate_only = self.get_parent_layer().model.evaluate_only  # type: ignore
            if not self.evaluate_only:
                raise NotImplementedError("The convDirect module only works in evaluate_only mode!")
        else:
            self.evaluate_only = True

        if ConvDirect.lib_cd is None:
            ConvDirect.lib_cd = load_library("convDirect")

        self._DT = ctypes.POINTER(ctypes.c_float)()
        self._FT = ctypes.POINTER(ctypes.c_float)()
        self._YT = ctypes.POINTER(ctypes.c_float)()

        try:
            [self._conv_direct_pre,
             self._conv_direct_kernel,
             self._conv_direct_post] = [getattr(self.__class__.lib_cd, method_name + suffix)
                                        for suffix in ('_pre', '_kernel', '_post')]
        except AttributeError as exc:
            raise NotImplementedError("Error: Method '{}' not supported by convDirect library.\n"
                                      "       Run convDirect_info to see the convDirect supported methods."
                                      "".format(method_name)) from exc

        self._reuse_processed_weights = False
        if self.evaluate_only and method_name.find("convdirect_block_blis") == 0:
            self._reuse_processed_weights = True
        self._weights_already_processed = False

    def encode_shape(self, shape):
        return encode_shape(shape, self.tensor_format)

    def decode_shape(self, shape):
        return decode_shape(shape, self.tensor_format)

    def conv_direct(self, weights: np.ndarray, x: np.ndarray,
                    # NOTE: "out" originally was called "biases"
                    out: np.ndarray | None = None,  # type: ignore
                    vpadding=0, hpadding=0,
                    vstride=1, hstride=1,
                    vdilation=1, hdilation=1,
                    relu=False, bn=False, running_mean=None,
                    inv_std=None, gamma=None, beta=None):

        n, ci, hi, wi = self.decode_shape(x.shape)

        if self.tensor_format is TensorFormat.NCHW:
            co, ci, kh, kw = weights.shape
        else:
            ci, kh, kw, co = weights.shape

        # TODO: Move to Conv2D_CPU (or remove)
        if out is None:
            ho = (hi + 2 * vpadding - vdilation * (kh - 1) - 1) // vstride + 1
            wo = (wi + 2 * hpadding - hdilation * (kw - 1) - 1) // hstride + 1
            bias_shape = self.encode_shape((n, co, ho, wo))
            out = np.zeros(bias_shape, weights.dtype, order="C")
        else:
            out = out[:n, :]

            match self.tensor_format:
                case TensorFormat.NCHW:
                    bb, knb, ho, wo = out.shape
                case TensorFormat.NHWC:
                    bb, ho, wo, knb = out.shape
                case tensor_format:
                    raise NotImplementedError(f"No support for {tensor_format} tensor format!")
            assert co == knb, "Number of filters must be the same!"
            assert n == bb, "Batch sizes must be the same!"
        out: np.ndarray

        # int t = n, Co = co, Ci = ci, Ho = h, Wo = w, Hf = r, Wf = s;
        (t, Co, Ci, Ho, Wo, Hf, Wf) = (n, co, ci, hi, wi, kh, kw)

        if self._reuse_processed_weights and self._weights_already_processed:
            self._DT = x.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
            self._YT = out.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
        else:
            # #define CONVDIRECT_PRE_PARAMS \
            #     int t, int Co, int Ci,    \
            #     int Ho, int Wo,           \
            #     int Hf, int Wf,           \
            #     const DTYPE *D,           \
            #     const DTYPE *F,           \
            #     const DTYPE *Y,           \
            #     DTYPE **DT,               \
            #     DTYPE **FT,               \
            #     DTYPE **YT
            self._conv_direct_pre(ctypes.c_int(t), ctypes.c_int(Co), ctypes.c_int(Ci),
                                  ctypes.c_int(Ho), ctypes.c_int(Wo),
                                  ctypes.c_int(Hf), ctypes.c_int(Wf),
                                  ctypes.c_void_p(x.ctypes.data),
                                  ctypes.c_void_p(weights.ctypes.data),
                                  ctypes.c_void_p(out.ctypes.data),
                                  ctypes.byref(self._DT),
                                  ctypes.byref(self._FT),
                                  ctypes.byref(self._YT))
            self._weights_already_processed = True

        # #define CONVDIRECT_KERNEL_PARAMS  \
        #     int t, int Co, int Ci,        \
        #     int Ho, int Wo,               \
        #     int Hf, int Wf,               \
        #     int vpadding, int hpadding,   \
        #     int vstride, int hstride,     \
        #     int vdilation, int hdilation, \
        #     DTYPE alpha,                  \
        #     const DTYPE *DT,              \
        #     const DTYPE *FT,              \
        #     DTYPE beta,                   \
        #     DTYPE *YT
        self._conv_direct_kernel(ctypes.c_int(t), ctypes.c_int(Co), ctypes.c_int(Ci),
                                 ctypes.c_int(Ho), ctypes.c_int(Wo),
                                 ctypes.c_int(Hf), ctypes.c_int(Wf),
                                 ctypes.c_int(vpadding), ctypes.c_int(hpadding),
                                 ctypes.c_int(vstride), ctypes.c_int(hstride),
                                 ctypes.c_int(vdilation), ctypes.c_int(hdilation),
                                 ctypes.c_float(1.0),
                                 self._DT,
                                 self._FT,
                                 ctypes.c_float(1.0),
                                 self._YT)

        if not self._reuse_processed_weights:
            # #define CONVDIRECT_POST_PARAMS \
            #     int t, int Co, int Ci,     \
            #     int Ho, int Wo,            \
            #     int Hf, int Wf,            \
            #     DTYPE **DT,                \
            #     DTYPE **FT,                \
            #     DTYPE **YT,                \
            #     DTYPE *Y
            self._conv_direct_post(ctypes.c_int(t), ctypes.c_int(Co), ctypes.c_int(Ci),
                                   ctypes.c_int(Ho), ctypes.c_int(Wo),
                                   ctypes.c_int(Hf), ctypes.c_int(Wf),
                                   ctypes.c_float(1.0),
                                   ctypes.byref(self._DT),
                                   ctypes.byref(self._FT),
                                   ctypes.byref(self._YT),
                                   ctypes.c_void_p(out.ctypes.data))
        return out


def time_it_func(x: np.ndarray, w_c: np.ndarray, out: np.ndarray,
                 b: int, kn: int,
                 ho: int, wo: int, kh: int, kw: int,
                 vpadding: int, hpadding: int, vstride: int, hstride: int,
                 vdilation: int, hdilation: int,
                 ) -> int | float:
    res = np.zeros(((x.shape[0] * ho * wo), (x.shape[-1] * kh * kw)), dtype=x.dtype)
    im2row_nhwc_cython(x, res,
                       kh, kw, ho, wo,
                       vpadding, hpadding, vstride, hstride,
                       vdilation, hdilation)
    res = res @ w_c
    res += out.reshape(b * ho * wo, kn)
    return res


def __usage_example__():
    # Imports for this usage example (not required otherwise)
    from timeit import timeit
    from pydtnn.utils import random
    # Default parameters (1st layer AlexNet for Cifar10)
    b = 32  # Batch size
    c = 16  # Channels per layer
    h = 18  # Layers height
    w = 18  # Layers width
    kn = 16  # Number of filters
    kh = 3  # Filters weights height
    kw = 3  # Filters weights width
    vpadding = 1  # Vertical padding
    hpadding = 1  # Horizontal padding
    vstride = 1  # Vertical stride
    hstride = 1  # Horizontal stride
    vdilation = 1  # Vertical dilation
    hdilation = 1  # Horizontal dilation
    # Create weights, x, and out matrices from previous parameters. If no out
    # matrix is provided, a proper one filled with zeros will be automatically
    # created.
    random.seed(0)
    # weights[1][1][1][1] = -322.0
    # weights[2][2][2][2] = -334.0

    ho = (h + 2 * vpadding - vdilation * (kh - 1) - 1) // vstride + 1
    wo = (w + 2 * hpadding - hdilation * (kw - 1) - 1) // hstride + 1

    # NCHW --------------------------
    weights = random.random((c, kh, kw, kn)).astype(np.float32, order='C')
    x = random.random((b, h, w, c)).astype(np.float32, order='C')
    out = (np.ones((b, ho, wo, kn)) * 10).astype(np.float32, order='C')

    print("Using conv_direct to compute weights * x + out...")
    # conv_direct = ConvDirect("convdirect_im2row_nhwc_default")
    conv_direct = ConvDirect("convdirect_block_blis_nhwc_blis")
    # conv_direct = ConvDirect("convdirect_conv_gemm_nhwc_default")
    # from ipdb import launch_ipdb_on_exception
    # with launch_ipdb_on_exception():
    conv_direct_result = conv_direct.conv_direct(weights, x, out,
                                                 vpadding=vpadding, hpadding=hpadding,
                                                 vstride=vstride, hstride=hstride,
                                                 vdilation=vdilation, hdilation=hdilation)
    conv_direct_t = timeit(lambda: conv_direct.conv_direct(weights, x, out,
                                                           vpadding=vpadding, hpadding=hpadding,
                                                           vstride=vstride, hstride=hstride,
                                                           vdilation=vdilation, hdilation=hdilation),
                           number=10) / 10
    print("Using im2col and mm NHWC ...")

    x_c = np.zeros(((x.shape[0] * ho * wo), (x.shape[-1] * kh * kw)), dtype=x.dtype)
    im2row_nhwc_cython(x, x_c,
                       kh, kw, ho, wo,
                       vpadding, hpadding, vstride, hstride, vdilation, hdilation)
    w_c = weights.reshape(-1, kn)
    im2col_mm_result_nhwc = (x_c @ w_c + out.reshape(b * ho * wo, kn)).reshape(-1, ho, wo, kn)
    mm_t = timeit(
        lambda: time_it_func(x, w_c, out, b, kn, ho, wo, kh, kw, vpadding, hpadding, vstride, hstride, vdilation, hdilation),
        number=10) / 10

    print("conv_direct time: {:.4f}".format(conv_direct_t))
    print("im2row + mm time: {:.4f}".format(mm_t))
    print("Sum WINOGRAD NHWC: ", conv_direct_result.sum(), conv_direct_result.shape)
    print("Sum   IM2COL NHWC: ", im2col_mm_result_nhwc.sum(), im2col_mm_result_nhwc.shape)
    print("np.allclose NHWC: ", np.allclose(conv_direct_result, im2col_mm_result_nhwc, atol=1e-3))


if __name__ == "__main__":
    __usage_example__()



import numpy as np

from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_EVENT_FINISHED, PYDTNN_OPS_EVENT_enum
from pydtnn.backends.gpu.layers.conv_2d import Conv2DGPU
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.utils.constants import DTYPE2CTYPE, ArrayShape


from pydtnn.utils.tensor import TensorFormat, format_transpose
from typing import Any, override

from pycuda.compiler import SourceModule  #type: ignore
from pycuda.driver import Function  #type: ignore

class Conv2DStandardGPU(Conv2DGPU):

    def _initializing_special_parameters(self):
        match self.model.tensor_format:
            case TensorFormat.NCHW:
                self.weights_shape = (self.co, self.ci, *self.filter_shape)
            case TensorFormat.NHWC:
                self.weights_shape = (self.ci, *self.filter_shape, self.co)
                # NOTE: It is this shape, even if in the CPU version is different.
                #self.weights_shape = (self.co, *self.filter_shape, self.ci)
            case _:
                raise NotImplementedError(f"\"{self.model.tensor_format}\" format not implemented.")
    # -----

    def initialize(self, prev_shape: ArrayShape, x: TensorGPU) -> None:
        super().initialize(prev_shape, x)

        self.dim_n = self.model.batch_size * self.ho * self.wo
        self.dim_c = self.ci * self.kh * self.kw

        match self.model.tensor_format:
            case TensorFormat.NCHW:
                im2_x_shape = (self.dim_c, self.dim_n)
                dw_shape = (self.co, self.dim_c)
                x_2im_var_shape = (self.dim_c, self.dim_n)

                self.im2_func = self.fwd_nchw(self.use_bias)
                self._2im_func = self._backward_nchw(self.use_bias)
            case TensorFormat.NHWC:
                im2_x_shape = (self.dim_n, self.dim_c)
                dw_shape = (self.dim_c, self.co)
                x_2im_var_shape = (self.dim_n, self.dim_c)

                self.im2_func = self.fwd_nhwc(self.use_bias)
                self._2im_func = self._backward_nhwc(self.use_bias)
            case _:
                raise NotImplementedError(f"\"{self.model.tensor_format}\" format not implemented.")

        self.im2_x = TensorGPU.create_zeros_tensor(im2_x_shape, self.model.dtype, self.model.tensor_format, self.model.cudnn_dtype)
        self.x_2im_var = TensorGPU.create_zeros_tensor(x_2im_var_shape, self.model.dtype, self.model.tensor_format, self.model.cudnn_dtype)
        
        self.y = TensorGPU.create_zeros_tensor((self.model.batch_size, *self.shape), self.model.dtype, self.model.tensor_format, self.model.cudnn_dtype)
        self.dw = TensorGPU.create_zeros_tensor(dw_shape, self.model.dtype, self.model.tensor_format, self.model.cudnn_dtype)
        self.dx = TensorGPU.create_zeros_tensor(self.x.ary.shape, self.model.dtype, self.model.tensor_format, self.model.cudnn_dtype)
    # -----

    @override
    def _export_weights_dw(self, key: str) -> Any:
        value = getattr(self, key)

        match self.model.tensor_format:
            case TensorFormat.NHWC:
                # NHWC's src: ci, kh, kw, co
                # NCHW's dst: co, ci, kh, kw
                gpu_ary = value.ary
                cpu_ary = gpu_ary.get()
                return np.asarray(format_transpose(cpu_ary, "IHWO", "OIHW"), dtype=np.float64, order="C", copy=True)
            case default:
                return super()._export_prop(key)
    # ------

    @override
    def _import_weights_dw(self, key: str, value: Any) -> None:
        attribute = getattr(self, key)
        match self.model.tensor_format:
            case TensorFormat.NHWC:
                # NCHW's src: co, ci, kh, kw
                # NHWC's dst: ci, kh, kw, co
                cpu_ary = np.asarray(format_transpose(value, "OIHW", "IHWO"), dtype=self.model.dtype, order="C", copy=None)
                attribute.ary.set(cpu_ary)
                return
            case default:
                return super()._import_prop(key, value)
    # ---

    def forward(self, x: TensorGPU) -> TensorGPU:
        # im2col / im2row
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CUDNN)
        self.im2_func(x.ary, self.weights.ary, 
                      self.im2_x.ary, self.y,
                      self.biases.ary,
                      np.int32(self.dim_c), np.int32(self.dim_n),
                      np.int32(self.model.batch_size), np.int32(self.ci), np.int32(self.hi), np.int32(self.wi),
                      np.int32(self.co), np.int32(self.ho), np.int32(self.wo),
                      np.int32(self.kh), np.int32(self.kw),
                      np.int32(self.vpadding), np.int32(self.hpadding),
                      np.int32(self.vstride), np.int32(self.hstride),
                      np.int32(self.vdilation), np.int32(self.hdilation),
                      grid = self.grid, block = self.block,
                      stream = self.model.stream
                    )
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return self.y
    # ---

    def backward(self, dy: TensorGPU) -> TensorGPU:
        
        self.dx.fill(0)

        # im2col / im2row
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_CUDNN_DX)
        self._2im_func(dy.ary,
                       self.im2_x.ary,
                       self.weights.ary,
                       self.dw.ary,
                       self.db.ary,
                       self.dx.ary,
                       self.x_2im_var.ary,
                       np.int32(self.dim_c), np.int32(self.dim_n),
                       np.int32(self.model.batch_size), np.int32(self.ci), np.int32(self.hi), np.int32(self.wi),
                       np.int32(self.co), np.int32(self.ho), np.int32(self.wo),
                       np.int32(self.kh), np.int32(self.kw),
                       np.int32(self.vpadding), np.int32(self.hpadding),
                       np.int32(self.vstride), np.int32(self.hstride),
                       np.int32(self.vdilation), np.int32(self.hdilation),
                       grid = self.grid, block = self.block,
                       stream = self.model.stream
                    )
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return self.dx
    # ---

#########################################################################################################
## CUDA CODE ##
###############
    #========================
    #= FORWARD-related code =
    #========================

    BIAS= \
"""
    for(i = idx; i < dim_n; i += num_workers)
        for(j = 0; j < co; j++)
    {{
        *(im2_var + SHIFT(i, j, co)) += (*(bias + j));
    }}
""" 
    # -- END BIAS --

    DB = \
"""
    for (ci = idx; ci < c; ci += num_workers)
    {{
        *(db + ci) = 0;
        for (ni = 0; ni < n; ni++)
            for (hi = 0; hi < h; hi++)
                for (wi = 0; wi < w; wi++)
        {{
            *(db + ci) += *(dy + SHIFT_DY(ni, ci, hi, wi, c, h, w));
        }}
    }}
"""
    # -- END DB --

    def fwd_nchw(self, use_bias: bool) -> Function:
        # im2_var.shape = (self.dim_c, self.dim_n) = (self.ci * self.kh * self.kw, self.model.batch_size * self.ho * self.wo)
        code = \
"""
// im2col-related macros
#define GET_CI(row, h, w) row / (w * h)
#define GET_KI(row, h, w) (row / w) % h
#define GET_KJ(row, h, w) row % w
#define IS_BETWEEN(min_v, var, max_v) (min_v <= var) && (var < max_v)
#define SHIFT_COLS(row, col, dim_cols) row * dim_cols + col
#define SHIFT_X(ni, ci, hi, wi, c, h, w) ((ni * c + ci) * h + hi) * w + wi

// matmul-related macros
#define SHIFT(i, j, dim_j) i * dim_j + j
#define GET_I(idx, dim_j) * idx / dim_j
#define GET_J(idx, dim_j) * idx % dim_j

__global__ void {FUNC_NAME}(const {T} *const x,
                            const {T} *const weights,
                            {T}* im2_var, {T}* y,
                            {T}* bias,
                            int dim_c, int dim_n,
                            int n, int c, int h, int w,
                            int co, int ho, int wo,
                            int kh, int kw, 
                            int vpadding, int hpadding,
                            int vstride, int hstride, 
                            int vdilation, int hdilation)
{{  
    const int idx = blockIdx.x * blockDim.x + threadIdx.x
    const int num_workers = blockDim.x * gridDim.x;
    
    // im2col const
    const int N = c * kh * kw;
    const int dim_cols = n * self.ho * self.wo;
    // matmul const
    const int N_MATMUL = co * dim_n;
    
    // im2col vars
    int ci, ki, kj, ni, hoi, hi, wi, woi, idx, row, col;
    // matmul vars
    int i, j, k;

    // Im2Col
    for(row = idx; row < N; row += num_workers)
    {{
        ci = GET_CI(row, h, w);
        ki = GET_KI(row, h, w);
        kj = GET_KJ(row, h, w);
        for (ni = 0; ni < n; ni++) for (hoi = 0; hoi < ho; hoi++)
        {{
            hi = vstride * hoi + vdilation * ki - vpadding;
            for (woi = 0; woi < wo; woi++)
            {{
                wi = hstride * woi + hdilation * kj - hpadding;
                col = (ni * ho + hoi) * wo + woi;
                //im2_var[row, col] = ((0 <= hi) && (hi < h) && (0 <= wi) && (wi < w)) ? x[nn, cc, x_x, x_y] : ({T}) 0.0;
                if (IS_BETWEEN(0, hi, h) && IS_BETWEEN(0, wi, w))
                    *(im2_var + SHIFT_COLS(row, col, dim_cols)) = *(x + SHIFT_X(n, ci, hi, wi, c, h, w));
                else
                    *(im2_var + SHIFT_COLS(row, col, dim_cols)) = ({T}) 0.0;
            }}
        }}
    }}
    __syncthreads();

    // Matmul - w_rows X x_cols = y.T
    // weights.shape "=" (co, dim_c); im2_var.shape = (dim_c, dim_n); y.T "="(co, dim_n); y.shape "=" (dim_n, co) || "=": because it's not equal, but "equivalent" in this situation.
    for(i_j = idx; i_j < N_MATMUL; i_j += num_workers)
    {{
        i = GET_I(i_j, dim_n);
        j = GET_J(i_j, dim_n);
        for(k = 0; k < dim_c; k++)
        {{
            // y[i, j] += weights[i, k] * im2_var[k, j]
            *(y + SHIFT(j, i, co)) += (*(weights + SHIFT(i, k, dim_c))) * (*(im2_var + SHIFT(k, j, dim_n)));
        }}
    }}
#if {USE_BIAS}

    __syncthreads();
    {BIAS}
#endif

}}
"""
        _t = DTYPE2CTYPE[self.model.dtype]  # variable Type

        func_name = "im2col_fwd_gpu"
        code = code.format(FUNC_NAME=func_name,
                           T=_t,
                           USE_BIAS=use_bias,
                           BIAS=self.BIAS
                           )
        module = SourceModule(code).get_function(func_name)

        return module
    # -------------------------

    def fwd_nhwc(self, use_bias:bool) -> Function:
        # cols.shape = (self.dim_n, self.dim_c) = (self.model.batch_size * self.ho * self.wo, self.ci * self.kh * self.kw)
        code = \
"""
#define GET_NI(row, h, w) row / (w * h)
#define GET_HO(row, h, w) (row / w) % h
#define GET_WO(row, h, w) row % w
#define IS_BETWEEN(min_v, var, max_v) (min_v <= var) && (var < max_v)
#define SHIFT_ROWS(row, col, dim_cols) row * dim_cols + col
// NOTE: This is NHWC
#define SHIFT_X(ni, ci, hi, wi, c, h, w) ((ni * h + hi) * w + wi) * c + ci

// matmul-related macros
#define SHIFT(i, j, dim_j) i * dim_j + j
#define GET_I(idx, dim_j) * idx / dim_j
#define GET_J(idx, dim_j) * idx % dim_j

__global__ void {FUNC_NAME}(const {T} *const x,
                            const {T} *const weights,
                            {T}* im2_var, {T}* y,
                            {T}* bias,
                            int dim_c, int dim_n,
                            int n, int c, int h, int w,
                            int co, int ho, int wo,
                            int kh, int kw, 
                            int vpadding, int hpadding,
                            int vstride, int hstride, 
                            int vdilation, int hdilation)
{{
    int ci, ki, kj, ni, hoi, hi, wi, woi, idx, row, col;
    int i, j, k, i_j;

    const int idx = blockIdx.x * blockDim.x + threadIdx.x
    const int num_workers = blockDim.x * gridDim.x;
    const int N = n * ho * wo;
    const int dim_cols = n * self.ho * self.wo;

    const int N_matmul = dim_n * co;

    // Im2Row
    for(row = idx; row < N; row += num_workers)
    {{  
        ni = GET_NI(row, n, ho, wo);
        hoi = GET_HO(row, n, ho, wo);
        woi = GET_WO(row, n, ho, wo);
        for (ci = 0; ci < c; ci++)
        {{
            for (ki = 0; ki < kh; ki++)
            {{
                hi = vstride * hoi + vdilation * ki - vpadding;
                for (kj = 0; kj < kw; kj++)
                {{
                    wi = hstride * woi + hdilation * kj - hpadding;
                    col = (ni * ho + hoi) * wo + woi;
                    
                    //im2_var[row, col] = ((0 <= hi) && (hi < h) && (0 <= wi) && (wi < w)) ? x[nn, cc, x_x, x_y] : ({T}) 0.0;
                    if (IS_BETWEEN(0, hi, h) && IS_BETWEEN(0, wi, w))
                        *(im2_var + SHIFT_ROWS(row, col, dim_cols)) = *(x + SHIFT_X(n, ci, hi, wi, c, h, w));
                    else
                        *(im2_var + SHIFT_ROWS(row, col, dim_cols)) = ({T}) 0.0;
                }}
            }}
        }}
    }}

    __syncthreads();

    // Matmul - im2_var X w_rows = y
    // im_var = (i, k)
    // w_rows = (k, j)
    // y = (i, j)

    // im2_var.shape = (dim_n, dim_c); weights.shape "=" (dim_c, co); y.shape "=" (dim_n * co) || "=": because it's not equal, but "equivalent" in this situation.    
    for(i_j = idx; i_j < N_matmul; i_j += num_workers)
        for(k = 0; k < dim_c; k++)
    {{  
        i = GET_I(i_j, co);
        j = GET_J(i_j, co);
        // y[i, j] += im2_var[i, k] * weights[k, j]
        *(y + i_j) += (*(im2_var + SHIFT(i, k, dim_c))) * (*(weights + SHIFT(k, j, co)));
    }}
#if {USE_BIAS}

    __syncthreads();
    {BIAS}
#endif

}}
"""
        _t = DTYPE2CTYPE[self.model.dtype]  # variable Type

        func_name = "im2row_fwd_gpu"
        code = code.format(FUNC_NAME=func_name,
                            T=_t,
                            USE_BIAS=use_bias,
                            BIAS=self.BIAS
                            )
        module = SourceModule(code).get_function(func_name)

        return module
    # -------------------------
    
    #=========================
    #= BACKWARD-related code =
    #=========================

    def _backward_nchw(self, use_bias:bool) -> Function:
        code = \
"""
#define IS_BETWEEN(min_v, var, max_v) (min_v <= var) && (var < max_v)

#define SHIFT_DY(ni, ci, hi, wi, n, c, h, w) (((((ni * c) + ci) * h) + hi) * w + wi)

// matmul-related macros
#define SHIFT(i, j, dim_j) i * dim_j + j
#define GET_I(idx, dim_j) * idx / dim_j
#define GET_J(idx, dim_j) * idx % dim_j

// im2col-related macros
#define GET_N(idx, n, c, h, w) idx / (w * h * c)
#define GET_C(idx, n, c, h, w) (idx / (w * h)) % c
#define GET_H(idx, n, c, h, w) (idx / w) % h
#define GET_W(idx, n, c, h, w) idx % w

__global__ void {FUNC_NAME}(const {T} *const dy,
                            const {T} *const im2_var,
                            const {T} *const weights,
                            {T}* dw, {T}* db, {T}* dx
                            {T}* col_2im_var,
                            int dim_c, int dim_n,
                            int n, int c, int h, int w,
                            int co, int ho, int wo,
                            int kh, int kw, 
                            int vpadding, int hpadding,
                            int vstride, int hstride, 
                            int vdilation, int hdilation)
{{
    // NOTE: c, h, w are the input ones and co, ho, wo are the output ones (they may differ)
    // base_dy.shape = (n, co, ho, wo)
    // im2_var.shape = (dim_n, co) || dim_n = (n * self.ho * self.wo)
    // weights.shape = (co, c, kh, kw);
    // dy.shape = (co, n, ho, wo)
    // dw.shape = (co, dim_c); dim_c = (c, kh, kw)
    // db.shape = (co, )
    // dx.shape = (n, c, h, w)
    // col_2im_var.shape = (dim_c, dim_n); dim_n = n * ho * wo; dim_c = c * kh * kw

    const int idx = blockIdx.x * blockDim.x + threadIdx.x
    const int num_workers = blockDim.x * gridDim.x;
    const int N_DW = co * c * kh * kw;
    const int N_COL2IM_VAR = dim_c * dim_n;
    const int N_COL2IM = n * c * h * w;
    const int N_TRANSPOSE = n * co * ho * wo;

    int i, j, k, i_j, dim_j, khi, kwi;
    int ni, ci, hi, wi, dy_i;

    // Matmul dy transposed and im2_var.T in and save it in dw
    // NOTE: Here dy is treated as (co, n*ho*wo); im2_var.T.shape = (n*ho*wo, ci*kh*kw)
    dim_j = N_DW / co;

    // NOTE: Remember -> dy base: NCHW, the dy needed to work: CNHW
    //dw.shape - (co, c, kh, kw)

    for(i_j = idx; i_j < N_DW; i_j += workers)
        for(k = 0; k < dim_n; k++)
    {{
        i = GET_I(i_j, dim_j);
        j = GET_J(i_j, dim_j);

        // Accessing dy like it was transposed from NCHW to CNHW
        //dy "=" (co, n, self.ho, self.wo)
        //i = "co" = GET_N(k, c, n, h, w).
        ni = GET_C(k, co, n, ho, wo);
        hi = GET_H(k, co, n, ho, wo);
        wi = GET_W(k, co, n, ho, wo);

        *(dw + i_j) += (*(dy + SHIFT_DY(ni, i, hi, wi, co, ho, wo))) * (*(im2_var + SHIFT(k, j, dim_j)));
    }}

    // np.sum(dy, axis=(0,2,3), out=db)
#if {USE_BIAS}
    {DB}
#endif

    // col_2im_var "=" (dim_c, dim_n) = (c * kh * kw, n * ho * wo)
    // mamtul(weights.reshape(co, -1).T, tranposed dy) ==>
    // mamtul(weights.reshape(co, c * kh * kw).T, tranposed dy) 
    // tranposed dy.shape = (co, n*ho*wo)
    for(i_j = idx; i_j < N_COL2IM_VAR; i_j += num_workers)
        for(k = 0; k < co; k++)
    {{
        i = GET_I(i_j, dim_n);
        j = GET_J(i_j, dim_n);

        // Accessing dy like it was transposed from NCHW to CNHW
        //dy "=" (co, n, self.ho, self.wo)
        dy_i = SHIFT(k, j, dim_n)
        ni = GET_C(dy_i, co, n, ho, wo);
        hi = GET_H(dy_i, co, n, ho, wo);
        wi = GET_W(dy_i, co, n, ho, wo);

        //col_2im_var[i][j] = weights[i][k] * dy[k][j]
        *(col_2im_var + i_j) =  (*(weights + SHIFT(i, k, co))) * (*(dy + SHIFT_DY(ni, k, hi, wi, co, ho, wo)));
    }}

    __syncthreads();

    // Col2Im
    for (i = idx; i < N_COL2IM; i += num_workers)
    {{
        ni = GET_N(i, n, c, h, w);
        ci = GET_C(i, n, c, h, w);
        hx = GET_H(i, n, c, h, w);
        wx = GET_W(i, n, c, h, w);

        for (khi = 0; khi < kh; khi++) 
            for (kwi = 0; kwi < kw; kwi++)
        {{
            // hx = vstride * xx + vdilation * khi - vpadding;
            xx = (hx + vpadding - vdilation * khi) / vstride;
            // wx = hstride * yy + hdilation * kwi - hpadding;
            yy = (wx + hpadding - hdilation * kwi) / hstride;

            x_o = (int) xx;
            y_o = (int) yy;
            
            // if (the variables have no decimals) and (are bewteen 0 and ho/wo):
            if ((x_o == xx) && (y_o == yy) && IS_BETWEEN(0, xx, ho) && IS_BETWEEN(0, yy, wo))
            {{
                row = cc * kh * kw + ii * kw + jj;
                col = nn * ho * wo + x_o * wo + y_o;
                //dx[nn, cc, x_x, x_y] += cols[row, col]
                *(dx + i) += (*(cols + SHIFT(row, col, dim_n)));
            }}
        }}
    }}

}}
"""
        _t = DTYPE2CTYPE[self.model.dtype]  # variable Type

        func_name = "col2im_bwd_gpu"
        code = code.format(FUNC_NAME=func_name,
                            T=_t,
                            USE_BIAS=use_bias, 
                            DB = self.DB
                            )
        module = SourceModule(code).get_function(func_name)

        return module
    # -------------------------

    def _backward_nhwc(self, use_bias:bool) -> Function:
        code = \
"""
#define IS_BETWEEN(min_v, var, max_v) (min_v <= var) && (var < max_v)
#define SHIFT_DY(ni, ci, hi, wi, c, h, w) ((ni * h + hi) * w + wi) * c + ci

// matmul-related macros
#define SHIFT(i, j, dim_j) i * dim_j + j
#define GET_I(idx, dim_j) * idx / dim_j
#define GET_J(idx, dim_j) * idx % dim_j

// im2col-related macros
#define GET_N(idx, n, c, h, w) idx / (c * w * h)
#define GET_H(idx, n, c, h, w) (idx / (c * w)) % h
#define GET_W(idx, n, c, h, w) (idx / c) % w
#define GET_C(idx, n, c, h, w) idx % c

__global__ void {FUNC_NAME}(const {T} *const dy,
                            const {T} *const im2_var,
                            const {T} *const weights,
                            {T}* dw, {T}* db, {T}* dx
                            {T}* row_2im_var,
                            int dim_c, int dim_n,
                            int n, int c, int h, int w,
                            int co, int ho, int wo,
                            int kh, int kw, 
                            int vpadding, int hpadding,
                            int vstride, int hstride, 
                            int vdilation, int hdilation)
{{
    const int idx = blockIdx.x * blockDim.x + threadIdx.x
    const int num_workers = blockDim.x * gridDim.x;
    const int N_DW = co * c * kh * kw;
    const int N_ROW2IM_VAR = dim_c * dim_n;
    const int N_ROW2IM = n * c * h * w;

    int i, j, k, i_j, dim_j, dim_k, khi, kwi;

    // NOTE: c, h, w are the input ones and co, ho, wo are the output ones (they may differ)
    // base_dy.shape = (n, co, ho, wo)
    // im2_var.shape = (dim_n, dim_c) || dim_n = (n * self.ho * self.wo)
    // weights.shape = (co, kh, kw, c)
    // dy.shape = (n, ho, wo, co)
    // dw.shape = (dim_c, co) || dim_c = (c, kh, kw)
    // db.shape = (co, )
    // dx.shape = (n, c, h, w)
    // row_2im_var.shape = (dim_c, dim_n) || dim_n = n * ho * wo; dim_c = c * kh * kw
    
    // dw = np.matmul(im2_var.T, dy.reshape(n*ho*wo, self.co)); im2_var.T.shape = (ci*kh*kw, n*ho*wo)
    for(i_j = idx; i_j < N_DW; i_j += workers)
        for(k = 0; k < dim_n; k++)
    {{
        i = GET_I(i_j, co);
        j = GET_J(i_j, co);
        *(dw + i_j) += (*(im2_var + SHIFT(k, i, dim_c))) * (*(dy + SHIFT(k, j, co)));
    }}

    // np.sum(dy, axis=(0,1,2), out=db)
#if {USE_BIAS}
    {DB}
#endif

    // row_2im_var "=" (dim_c, dim_n)
    //mamtul(weights.reshape(self.ci * self.kh * self.kw, co), tranposed dy) <== mamtul(weights.reshape(co, -1).T, tranposed dy)
    // tranposed dy.shape = (co, n*ho*wo)
    for(i_j = idx; i_j < N_ROW2IM_VAR; i_j += num_workers)
        for(k = 0; k < co; k++)
    {{
        i = GET_I(i_j, dim_n);
        j = GET_J(i_j, dim_n);

        //row_2im_var[i][j] += dy[i][k] * weights[k][j]; (weights= weights.reshape((-1, co)).T)
        *(row_2im_var + i_j) += (*(dy + SHIFT(i, k, co))) * (*(weights + SHIFT(j, k, co)));
    }}

    __syncthreads();

    // Row2Im
    for (i = idx; i < N_ROW2IM; i += num_workers)
    {{
        ni = GET_N(i, n, c, h, w);
        ci = GET_C(i, n, c, h, w);
        hx = GET_H(i, n, c, h, w);
        wx = GET_W(i, n, c, h, w);

        for (khi = 0; khi < kh; khi++) 
            for (kwi = 0; kwi < kw; kwi++)
        {{
            // hx = vstride * xx + vdilation * khi - vpadding;
            xx = (hx + vpadding - vdilation * khi) / vstride;
            // wx = hstride * yy + hdilation * kwi - hpadding;
            yy = (wx + hpadding - hdilation * kwi) / hstride;

            x_o = (int) xx;
            y_o = (int) yy;
            
            // if (the variables have no decimals) and (are bewteen 0 and ho/wo):
            if ((x_o == xx) && (y_o == yy) && IS_BETWEEN(0, xx, ho) && IS_BETWEEN(0, yy, wo))
            {{
                row = nn * ho * wo + x_o * wo + y_o;
                col = cc * kh * kw + ii * kw + jj;
                //dx[nn, x_x, x_y, cc] += rows[row, col]
                *(dx + i) += (*(row_2im_var + SHIFT(row, col, dim_c)));
            }}
        }}
    }}

}}
"""
        _t = DTYPE2CTYPE[self.model.dtype]  # variable Type

        func_name = "row2im_fwd_gpu"
        code = code.format(FUNC_NAME=func_name,
                            T=_t,
                            USE_BIAS=use_bias,
                            DB=self.DB
                            )
        module = SourceModule(code).get_function(func_name)

        return module
    # -------------------------

#########################################################################################################



from pydtnn.backends.cpu.utils.bn_inference_cython import bn_relu_inference_cython
from pydtnn.backends.cpu.layers.batch_normalization import BatchNormalizationCPU
from pydtnn.layers.batch_normalization_relu import BatchNormalizationRelu
from pydtnn.utils.constants import ArrayShape

import numpy as np


class BatchNormalizationReluCPU(BatchNormalizationRelu[np.ndarray], BatchNormalizationCPU):

    # NOTE: The "__init__" method is being made (more or less) in Model (in _apply_layer_fusion) and in FusedLayerMixIn.

    def initialize(self, prev_shape: ArrayShape, x: np.ndarray | None = None):
        super().initialize(prev_shape, x)

        self.inv_std = BatchNormalizationCPU.get_inv_std(self.running_var, self.epsilon, self.model.dtype)

        # NOTE: This attribute only stores data, its value before the operation doesn't matter; it's initalized due avoid warnings in "LayerAndActivationBase.export".
        self.y: np.ndarray = np.zeros(shape=(self.model.batch_size, *self.shape), dtype=self.model.dtype, order="C")
        self.forward = self._forward
        self.backward = self._backward

    def _forward(self, x: np.ndarray) -> np.ndarray:
        """Version of the forward function that uses the BN + Relu"""

        n = x.shape[0]
        if self.spatial:
            x = x.reshape((-1, self.ci), copy=False, order="C")

        y: np.ndarray = self.y[:n, :]
        bn_relu_inference_cython(x,
                                 y.reshape((-1, self.ci), copy=False, order="C"),
                                 self.running_mean,
                                 self.inv_std,
                                 self.gamma,
                                 self.beta)

        if self.spatial:
            y_shape = self.model.encode_shape((n, self.ci, self.hi, self.wi))
            y = y.reshape(y_shape, copy=False)

        return np.asarray(y, dtype=self.model.dtype, order='C', copy=None)

    def _backward(self, dy: np.ndarray) -> np.ndarray:
        raise NotImplementedError("Use a real backwards variant!")



import numpy as np

from pydtnn.backends.cpu.layers.abstract.block_layer import AbstractBlockLayerCPU
from pydtnn.layers.fc import FC
from pydtnn.layers.dropout import Dropout
from pydtnn.layers.multiplication import Multiplication
from pydtnn.layers.scalar import Scalar
from pydtnn.activations.softmax import Softmax

from pydtnn.layers.multi_head_attention import MultiHeadAttention
from pydtnn.model import Model
from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_OPS_EVENT_enum


class MultiHeadAttentionCPU(MultiHeadAttention[np.ndarray], AbstractBlockLayerCPU):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.FC_q = FC(shape=(self.heads*self.d_k,))  # Dim: embedl x heads*d_k
        self.FC_k = FC(shape=(self.heads*self.d_k,))  # Dim: embedl x heads*d_k
        self.FC_v = FC(shape=(self.heads*self.d_k,))  # Dim: embedl x heads*d_k
        self.FC_o = FC(shape=(self.embedl,))   # Dim: heads*d_k x embedl
        self.mult_qkt = Multiplication()
        self.scalar_dk = Scalar(1.0/np.sqrt(self.d_k))
        self.softmax = Softmax()
        self.dropout = Dropout(rate=self.dropout_rate)
        self.mult_smv = Multiplication()
        self.mult_o = Multiplication()
        self.paths = [[self.FC_q, self.FC_k, self.FC_v, self.mult_qkt, self.scalar_dk, self.FC_o,
                       self.softmax, self.dropout, self.mult_smv, self.mult_o]]

        # The next attributes will be initialized later
        self.mask = None

    def initialize(self, prev_shape, x):
        super().initialize(prev_shape, x)
        self.shape = prev_shape
        self.embedl = prev_shape[-1]
        seq = prev_shape[-2]

        # Initialize all sublayers
        for layer in self.children:
            layer.init_backend_from_model(self.model)

        self.FC_q.initialize(prev_shape=(self.embedl,), x=x)
        self.FC_k.initialize(prev_shape=(self.embedl,), x=self.FC_q.y)
        self.FC_v.initialize(prev_shape=(self.embedl,), x=self.FC_k.y)
        self.mult_qkt.initialize(prev_shape=(1,), x=self.FC_v.y)
        self.scalar_dk.initialize(prev_shape=(1,), x=self.mult_qkt.y)
        self.FC_o.initialize(prev_shape=(self.heads*self.d_k,), x=self.scalar_dk.y)
        self.softmax.initialize(prev_shape=(self.heads, seq, seq,), x=self.FC_o.y)
        self.dropout.initialize(prev_shape=(self.heads, seq, seq,), x=self.softmax.y)
        self.mult_smv.initialize(prev_shape=(1,), x=self.dropout.y)
        self.mult_o.initialize(prev_shape=(1,), x=self.mult_smv.y)

        for layer in self.children:
            if layer.fwd_time is not None:
                self.fwd_time += layer.fwd_time
            if layer.bwd_time is not None:
                self.bwd_time += layer.bwd_time
            if layer.nparams is not None:
                self.nparams += layer.nparams

    def initialize_block_layer(self):
        pass

    def transformation_addheads(self, x):
        return x.reshape((x.shape[:-1] + (self.heads, self.d_k))).swapaxes(-3, -2)

    def transformation_removeheads(self, x):
        return x.swapaxes(-3, -2).reshape((x.shape[:-3] + (x.shape[-2], self.heads*self.d_k)))

    def mask_apply(self, x, mask):
        if len(mask.shape) == 2:
            seq, seq2 = mask.shape
        else:
            _, seq, seq2 = mask.shape
        if seq == 1:
            for j in range(self.heads):
                for k in range(seq2):
                    x[:, j, k] = x[:, j, k] * mask[:, 0]
        else:
            for j in range(self.heads):
                x[:, j] = x[:, j] * mask[:]
        return x

    def transpose(self, x):
        return x.swapaxes(-2, -1)

    def forward(self, query, key, value, mask=None):
        if self.model.mode == Model.Mode.TRAIN:
            self.mask = mask

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_MHA_FC_QKV)
        query = self.transformation_addheads(self.FC_q.forward(query))
        key = self.transformation_addheads(self.FC_k.forward(key))
        value = self.transformation_addheads(self.FC_v.forward(value))
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, 0)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_MHA_MATMUL_QK)
        score = self.mult_qkt.forward(query, self.transpose(key))
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, 0)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_MHA_SCALARDK)
        score = self.scalar_dk.forward(score)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, 0)

        if self.mask is not None:
            score = self.mask_apply(score, mask)
        score = self.softmax.forward(score)
        score = self.dropout.forward(score)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_MHA_MATMUL_SMV)
        score = self.mult_smv.forward(score, value)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, 0)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_MHA_FC_O)
        score = self.transformation_removeheads(score)
        score = self.FC_o.forward(score)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, 0)

        return score

    def backward(self, dy):
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_MHA_FC_O)
        dx = self.FC_o.backward(dy)
        dx = self.transformation_addheads(dx)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, 0)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_MHA_MATMUL_SMV)
        dx, d_value = self.mult_smv.backward(dx)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, 0)

        dx = self.dropout.backward(dx)
        dx = self.softmax.backward(dx)
        if self.mask is not None:
            dx = self.mask_apply(dx, self.mask)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_MHA_SCALARDK)
        dx = self.scalar_dk.backward(dx)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, 0)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_MHA_MATMUL_QK)
        d_query, d_key = self.mult_qkt.backward(dx)
        d_key = self.transpose(d_key)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, 0)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_MHA_FC_QKV)
        d_query = self.FC_q.backward(self.transformation_removeheads(d_query))
        d_key = self.FC_k.backward(self.transformation_removeheads(d_key))
        d_value = self.FC_v.backward(self.transformation_removeheads(d_value))
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, 0)

        return d_query, d_key, d_value



from pydtnn.layer_base import FusedLayerMixIn
from pydtnn.layers.conv_2d import Conv2D
from pydtnn.layers.batch_normalization import BatchNormalization
from pydtnn.utils.constants import Array

class Conv2DBatchNormalization[T: Array](FusedLayerMixIn[T], Conv2D[T], BatchNormalization[T]):
    pass



from pydtnn.metrics.metric import Metric

from numpy import ndarray


class MetricCPU(Metric[ndarray]):
    """
    Extends a Metric class with the attributes and methods required by CPU Metrics.
    """



from pydtnn.backends.cpu.layers.layer import LayerCPU
from pydtnn.layers.input import Input
import numpy as np


class InputCPU(Input[np.ndarray], LayerCPU):

    def forward(self, x: np.ndarray) -> np.ndarray:
        return np.asarray(x, dtype=self.model.dtype, order="C", copy=None)

    def backward(self, dy: np.ndarray) -> np.ndarray:
        return np.asarray(dy, dtype=self.model.dtype, order="C", copy=None)



import numpy as np

from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_EVENT_FINISHED, PYDTNN_OPS_EVENT_enum
from pydtnn.backends.gpu.layers.conv_2d import Conv2DGPU, MACROS_NCHW, MACROS_NHWC
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.utils.constants import ArrayShape, DTYPE2CTYPE

from pydtnn.utils.tensor import TensorFormat
from typing import Any, override

import pycuda.gpuarray as gpuarray  #type: ignore
from pycuda.compiler import SourceModule  #type: ignore
from pycuda.driver import Function  #type: ignore

class Conv2DDepthwiseGPU(Conv2DGPU):

    def _initializing_special_parameters(self):
        # Setting other parameters
        self.co = self.ci
        # Setting weights
        self.weights_shape = (1, self.ci, *self.filter_shape)

    def initialize(self, prev_shape: ArrayShape, x: TensorGPU) -> None:
        super().initialize(prev_shape, x)

        func_name: str = ""
        macros: str = ""
        self.bias_sum_bwd: Function = None

        match self.model.tensor_format:
            case TensorFormat.NCHW:
                func_name = "cuda_depthwise_conv_2d_{fwd_bwd}_nchw"
                macros = MACROS_NCHW
                self.bias_sum_bwd = self.cuda_sum_bias_axis_023()
                self.forward = self._forward_depthwise_nchw
                self.backward = self._backward_depthwise_nchw
            case TensorFormat.NHWC:
                func_name = "cuda_depthwise_conv_2d_{fwd_bwd}_nhwc"
                macros = MACROS_NHWC
                self.bias_sum_bwd = self.cuda_sum_bias_axis_012()
                self.forward = self._forward_depthwise_nhwc
                self.backward = self._backward_depthwise_nhwc
            case _:
                # TODO: self devolvía la versión con el número
                raise NotImplementedError(f"\"{self.name}\" is not implemented for \"{self.model.tensor_format}\" format.")

        self.total_num_threads = np.prod(self.grid, dtype=np.int32) * np.prod(self.block, dtype=np.int32)

        y_gpu = gpuarray.zeros((self.model.batch_size, *self.shape), self.model.dtype)
        self.y = TensorGPU(y_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        dx_gpu = gpuarray.zeros((self.model.batch_size, *self.shape), self.model.dtype)
        self.dx = TensorGPU(dx_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        self.fwd_func: Function = self.cuda_depthwise_conv_2d_fwd(func_name.format(fwd_bwd="fwd"), macros)
        self.bwd_func: Function = self.cuda_depthwise_conv_2d_bwd(func_name.format(fwd_bwd="bwd"), macros)
        self.bias_sum_fwd: Function = self.cuda_bias_sum_fwd_depthwise_conv()
    # ----

    def _forward_depthwise_nchw(self, x: TensorGPU) -> TensorGPU:
        self.x = x
        self.y.fill(0)

        n, c, h, w = x.shape

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CUDNN)
        self.fwd_func(x.ary, self.weights.ary, self.y.ary,
                      np.int32(self.vpadding), np.int32(self.hpadding),
                      np.int32(self.vstride), np.int32(self.hstride),
                      np.int32(self.vdilation), np.int32(self.hdilation),
                      np.int32(n), np.int32(c), np.int32(h), np.int32(w),
                      np.int32(self.kh), np.int32(self.kw), np.int32(self.ho), np.int32(self.wo),
                      self.total_num_threads, grid=self.grid, block=self.block,
                      stream=self.model.stream)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        if self.use_bias:
            self.biases: TensorGPU
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CUDNN_SUM_BIASES)
            self.bias_sum_fwd(x.ary, self.biases.ary,
                              np.int32(n), np.int32(c), np.int32(h), np.int32(w),
                              np.int32(n * h * w * c),
                              self.total_num_threads,
                              grid=self.grid, block=self.block,
                              stream=self.model.stream)
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return self.y
    # ----

    def _forward_depthwise_nhwc(self, x: TensorGPU) -> TensorGPU:
        self.x = x
        n, h, w, c = x.shape
        self.y.fill(0)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CUDNN)
        self.fwd_func(x.ary, self.weights.ary, self.y.ary,
                      np.int32(self.vpadding), np.int32(self.hpadding),
                      np.int32(self.vstride), np.int32(self.hstride),
                      np.int32(self.vdilation), np.int32(self.hdilation),
                      np.int32(n), np.int32(c), np.int32(h), np.int32(w),
                      np.int32(self.kh), np.int32(self.kw), np.int32(self.ho), np.int32(self.wo),
                      self.total_num_threads, grid=self.grid, block=self.block,
                      stream=self.model.stream)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        if self.use_bias:
            self.biases: TensorGPU
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CUDNN_SUM_BIASES)
            self.bias_sum_fwd(x.ary, self.biases.ary,
                              np.int32(n), np.int32(c), np.int32(h), np.int32(w),
                              np.int32(n * h * w * c),
                              self.total_num_threads,
                              grid=self.grid, block=self.block,
                              stream=self.model.stream)
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return self.y
    # ----

    def _backward_depthwise_nchw(self, dy: TensorGPU) -> TensorGPU:

        n, c, h, w = dy.shape
        self.dx.fill(0)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_CUDNN_DX)
        self.fwd_func(dy.ary, self.x.ary, self.weights.ary,
                      self.dx.ary, self.dw.ary,
                      np.int32(self.vpadding), np.int32(self.hpadding),
                      np.int32(self.vstride), np.int32(self.hstride),
                      np.int32(self.vdilation), np.int32(self.hdilation),
                      np.int32(n), np.int32(c), np.int32(h), np.int32(w),
                      np.int32(self.kh), np.int32(self.kw), np.int32(self.ho), np.int32(self.wo),
                      self.total_num_threads,
                      grid=self.grid, block=self.block, stream=self.model.stream)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        if self.use_bias:
            self.biases: TensorGPU
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_CUDNN_DB)
            self.bias_sum_bwd(dy.ary, self.db.ary,
                              np.int32(c), np.int32(h), np.int32(w),
                              np.int32(n * c * h * w), self.total_num_threads,
                              grid=self.grid, block=self.block, stream=self.model.stream)
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return self.dx
    # -----

    def _backward_depthwise_nhwc(self, dy: TensorGPU) -> TensorGPU:
        n, h, w, c = dy.shape
        self.dx.fill(0)

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_CUDNN_DX)
        self.fwd_func(dy.ary, self.x.ary, self.weights.ary,
                      self.dx.ary, self.dw.ary,
                      np.int32(self.vpadding), np.int32(self.hpadding),
                      np.int32(self.vstride), np.int32(self.hstride),
                      np.int32(self.vdilation), np.int32(self.hdilation),
                      np.int32(n), np.int32(c), np.int32(h), np.int32(w),
                      np.int32(self.kh), np.int32(self.kw), np.int32(self.ho), np.int32(self.wo),
                      self.total_num_threads, grid=self.grid, block=self.block,
                      stream=self.model.stream)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        if self.use_bias:
            self.biases: TensorGPU
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_CUDNN_DB)
            self.bias_sum_bwd(dy.ary, self.db.ary,
                              np.int32(c), np.int32(n * h * w * c),
                              self.total_num_threads,
                              grid=self.grid, block=self.block, stream=self.model.stream)
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return self.dx
    # -----

    @override
    def _export_weights_dw(self, key: str) -> Any:
        value = getattr(self, key)

        match self.model.tensor_format:
            case TensorFormat.NHWC:
                gpu_ary = value.ary
                cpu_ary = np.squeeze(gpu_ary.get(), axis=0)
                return np.asarray(cpu_ary, dtype=np.float64, order="C", copy=True)
            case TensorFormat.NCHW:
                gpu_ary = value.ary
                cpu_ary = np.squeeze(gpu_ary.get(), axis=0)
                return np.asarray(cpu_ary, dtype=np.float64, order="C", copy=True)
            case default:
                return super()._export_prop(key)
    # ---

    @override
    def _import_weights_dw(self, key: str, value: Any) -> None:
        attribute = getattr(self, key)
        match self.model.tensor_format:
            case TensorFormat.NHWC:
                cpu_ary = np.asarray(value, dtype=self.model.dtype, order="C", copy=None)
                cpu_ary = np.expand_dims(cpu_ary, axis=0)
                attribute.ary.set(cpu_ary)
                return
            case TensorFormat.NCHW:
                gpu_ary = attribute.ary
                cpu_ary = np.asarray(np.expand_dims(value, axis=0), dtype=self.model.dtype, order="C", copy=None)
                gpu_ary.set(cpu_ary)
                return 
            case default:
                return super()._import_prop(key, value)
    # ---


#########################################################################################################
## CUDA CODE ##
###############

    def cuda_depthwise_conv_2d_fwd(self, _func_name: str, _macros: str) -> Function:

        code = \
"""
{macros}
__global__ void {func_name}({T}* x, {T}* k, {T}* res,
                            int vpadding, int hpadding,
                            int vstride, int hstride,
                            int vdilation, int hdilation,
                            int n, int c, int h, int w,
                            int kh, int kw, int ho, int wo,
                            int num_workers)
{{
    int idx, cc, hi, wi, yy, xx, nn, x_x, x_y;
    int N = n * c * ho * wo;
    {T} val_k, val_x;

    for(idx = blockIdx.x * blockDim.x + threadIdx.x; idx < N; idx += num_workers)
    {{
        cc = INDEX_C(idx, c, ho, wo);
        xx = INDEX_H(idx, c, ho, wo);
        yy = INDEX_W(idx, c, ho, wo);

        for (hi = 0; hi < kh; hi++)
        {{
            for (wi = 0; wi < kw; wi++)
            {{
                x_x = vstride * xx + vdilation * hi - vpadding;
                x_y = hstride * yy + hdilation * wi - hpadding;
                if ((0 <= x_x) && (x_x < h) && (0 <= x_y) && (x_y < w))
                {{
                    val_k = *(SHIFT_POINTER(k, c, h, w, 0, cc, hi, wi));
                    val_x = *(SHIFT_POINTER(x, c, h, w, nn, cc, x_x, x_y));
                    *(SHIFT_POINTER(res, c, h, w, nn, cc, xx, yy)) += ({T}) (val_k * val_x);
                }}
            }}
        }}
    }}
}}
"""
        _t = DTYPE2CTYPE[self.model.dtype]  # variable Type

        code = code.format(macros=_macros,
                           func_name=_func_name,
                           T=_t
                           )
        module = SourceModule(code).get_function(_func_name)

        return module
    # ---

    def cuda_depthwise_conv_2d_bwd(self, _func_name: str, _macros: str) -> Function:

        code = \
"""
{macros}
__global__ void {func_name}({T}* dy, {T}* x, {T}* k,
                            {T}* dx, {T}* dw,
                            int vpadding, int hpadding,
                            int vstride, int hstride,
                            int vdilation, int hdilation,
                            int n, int c, int h, int w,
                            int kh, int kw, int ho, int wo,
                            int num_workers)
{{
    int idx, cc, khi, kwi, yy, xx, nn, x_x, x_y;
    {T} val_k, val_dy, val_x;
    int N = n * c * ho * wo;

    for(idx = blockIdx.x * blockDim.x + threadIdx.x; idx < N; idx += num_workers)
    {{
        cc = INDEX_C(idx, c, ho, wo);
        xx = INDEX_H(idx, c, ho, wo);
        yy = INDEX_W(idx, c, ho, wo);

        val_dy = ({T}) *(SHIFT_POINTER(dy, c, h, w, nn, cc, xx, yy));
        for (khi = 0; khi < kh; khi++)
        {{
            for (kwi = 0; kwi < kw; kwi++)
            {{
                x_x = vstride * xx + vdilation * khi - vpadding;
                x_y = hstride * yy + hdilation * kwi - hpadding;
                if ((0 <= x_x) && (x_x < h) && (0 <= x_y) && (x_y < w)){{
                    val_k = *(SHIFT_POINTER(k, c, h, w, 0, cc, khi, kwi));
                    val_x = *(SHIFT_POINTER(x, c, h, w, nn, cc, x_x, x_y));
                    *(SHIFT_POINTER(dw, c, h, w, 0, cc, khi, kwi)) = ({T}) (val_x * val_dy);
                    *(SHIFT_POINTER(dx, c, h, w, nn, cc, x_x, x_y)) += ({T}) (val_k * val_dy);
                }}
            }}
        }}
    }}
}}
"""
        _t = DTYPE2CTYPE[self.model.dtype]  # variable Type

        code = code.format(macros=_macros,
                           func_name=_func_name,
                           T=_t
                           )
        module = SourceModule(code).get_function(_func_name)

        return module
    # ---

    def cuda_bias_sum_fwd_depthwise_conv(self, _func_name: str = "bias_sum_fwd_depthwise_conv") -> Function:
        _t = DTYPE2CTYPE[self.model.dtype]  # variable Type

        code = \
"""
__global__ void {func_name}({T}* x, {T}* bias,
                            int co, int N,
                            int num_workers)
{{
    int idx;

    for(idx = blockIdx.x * blockDim.x + threadIdx.x; idx < N; idx += num_workers)
    {{
        *(x + idx) += *(bias + ( idx / (N/co) ) );
    }}
}}
"""

        code = code.format(func_name=_func_name,
                           T=_t
                           )
        module = SourceModule(code).get_function(_func_name)

        return module
    # ----
#########################################################################################################



import numpy as np

from pydtnn.metrics.categorical_hinge import CategoricalHinge

from pydtnn.backends.gpu.metrics.metric import MetricGPU
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pycuda.compiler import SourceModule  # type: ignore
from pycuda.driver import Function  # type: ignore

from pydtnn.utils.constants import DTYPE2CTYPE

class CategoricalHingeGPU(CategoricalHinge[TensorGPU], MetricGPU):

    def initialize(self) -> None:
        super().initialize()
        self.res = TensorGPU.create_zeros_tensor(shape=(1, ), dtype=np.dtype(self.model.dtype),
                                            tensor_format=self.model.tensor_format, cudnn_dtype=self.model.cudnn_dtype)
        self.local_res = TensorGPU.create_zeros_tensor(shape=(self.model.batch_size, ), dtype=np.dtype(self.model.dtype),
                                                  tensor_format=self.model.tensor_format, cudnn_dtype=self.model.cudnn_dtype)

    def __init_gpu_kernel__(self) -> Function:
        _name = "categorical_hinge"
        code = """

        #define SHIFT_2D_AR(p, i, j, dim_j) (p + ((i * dim_j) + j))

        __global__ void {name} ({T} *y_targ, {T} *y_pred, {T} *res, {T} *local_res, int n, int labels)
        {{
            int i, idx;
            {T} pos, neg, max_v, val_targ, val_pred;
        
            int base_idx = blockIdx.x * blockDim.x + threadIdx.x;
            int workers = blockDim.x * gridDim.x;

            for(idx = base_idx; idx < n; idx += workers)
            {{
                for(i = 0, max_v = ({T}) 0.0; i < labels; i++)
                {{
                    // val_targ = y_targ[idx][i];
                    val_targ = (*SHIFT_2D_AR(y_targ, idx, i, labels));

                    // val_pred = y_pred[idx][i];
                    val_pred = (*SHIFT_2D_AR(y_pred, idx, i, labels));
                    
                    pos += ({T}) (val_targ * val_pred);
                    neg = ({T}) (-1 * val_targ) + 1;
                    if ( (i == 0) || (max_v < neg))
                        max_v = neg;
                }}
                max_v = ({T}) ((max_v - pos) + 1);
                *(local_res + idx) = ({T}) (max_v > 0 ? max_v : 0);
            }}
            
            
            if(base_idx == 0)
            {{
                for(idx = 1; idx < n; idx++)
                    *(local_res) += *(local_res + idx);

                *(res) = ({T}) (*(local_res) / (n * labels));
            }}
        }}
        """.format(T=DTYPE2CTYPE[self.model.dtype],
                   name=_name)
        
        module = SourceModule(code).get_function(_name)
        return module

    def compute(self, y_pred: TensorGPU, y_targ: TensorGPU) -> float:
        n = y_pred.shape[0]

        self.res.fill(0)
        self.local_res.fill(0)

        n = np.int32(n)
        num_classes = np.int32(y_pred.shape[1])

        self.kernel(y_targ.ary, y_pred.ary, 
                    self.res.ary, self.local_res.ary,
                    n, num_classes,
                    grid=self.grid, block=self.block,
                    stream=self.model.stream)
        return self.res.ary.get()[0]



import numpy as _np
type _npDT = _np.int8 | _np.float32 | _np.float64
type _npDT_4Dims[T] = _np.ndarray[tuple[int, int, int, int], T]
type _npDT_4Dims[T] = _np.ndarray[tuple[int, int, int], T]
type _npDT_2Dims[T] = _np.ndarray[tuple[int, int], T]
type _npDT_1Dims[T] = _np.ndarray[tuple[int], T]

def transpose_0231_ikj_cython[T](original: _npDT_4Dims[T], transposed: _npDT_4Dims[T]) -> None:
    """
    Transposes a 4D matrix from (0,1,2,3) to (0,2,3,1).
    This is equivalent to transpose a 3D matrix 0x1x2·3 to 0x2·3x1

    Args:
        original npDT_3Dims): The original matrix.
        transposed npDT_3Dims): The matrix to transpose.
    Returns:
        Nothing. The output is stored in "transposed"
    """


def transpose_0231_ijk_cython[T](original: _npDT_4Dims[T], transposed: _npDT_4Dims[T]) -> None:
    """
    Transposes a 4D matrix from (0,1,2,3) to (0,2,3,1).
    This is equivalent to transpose a 3D matrix 0x1x2·3 to 0x2·3x1

    Args:
        original npDT_3Dims): The original matrix.
        transposed npDT_3Dims): The matrix to transpose.
    Returns:
        Nothing. The output is stored in "transposed"
    """

def transpose_0312_ikj_cython[T](original: _npDT_4Dims[T], transposed: _npDT_4Dims[T]) -> None:
    """
    Transposes a 4D matrix from (0,1,2,3) to (0,3,1,2).
    This is equivalent to transpose a 3D matrix 0x1·2x3 to 0x3x1·2

    Args:
        original npDT_3Dims): The original matrix.
        transposed npDT_3Dims): The matrix to transpose.
    Returns:
        Nothing. The output is stored in "transposed"
    """

def transpose_0312_ijk_cython[T](original: _npDT_4Dims[T], transposed: _npDT_4Dims[T]) -> None:
    """
    Transposes a 4D matrix from (0,1,2,3) to (0,3,1,2).
    This is equivalent to transpose a 3D matrix 0x1·2x3 to 0x3x1·2

    Args:
        original npDT_3Dims): The original matrix.
        transposed npDT_3Dims): The matrix to transpose.
    Returns:
        Nothing. The output is stored in "transposed"
    """


def transpose_1023_jik_cython[T](original: _npDT_4Dims[T], transposed: _npDT_4Dims[T]) -> None:
    """
    Transposes a 4D matrix from (0,1,2,3) to (1,0,2,3).
    This is equivalent to transpose a 3D matrix 0x1x2·3 to 1x0x2·3

    Args:
        original npDT_3Dims): The original matrix.
        transposed npDT_3Dims): The matrix to transpose.
    Returns:
        Nothing. The output is stored in "transposed"
    """


def transpose_1023_ijk_cython[T](original: _npDT_4Dims[T], transposed: _npDT_4Dims[T]) -> None:
    """
    Transposes a 4D matrix from (0,1,2,3) to (0,2,3,1).
    This is equivalent to transpose a 3D matrix 0x1x2·3 to 1x0x2·3

    Args:
        original npDT_3Dims): The original matrix.
        transposed npDT_3Dims): The matrix to transpose.
    Returns:
        Nothing. The output is stored in "transposed"
    """



from typing import TYPE_CHECKING

import numpy as np

from pydtnn.optimizers.optimizer import Optimizer
from pydtnn.utils.constants import Array

if TYPE_CHECKING:
    from pydtnn.model import Model


class RMSProp[T: Array](Optimizer[T]):
    """
    RMSProp optimizer
    """

    def __init__(self, learning_rate: float = 1e-2, rho: float = 0.9, epsilon: float = 1e-7,
                 decay: float = 0.0, dtype: np.dtype = np.dtype(np.float32)):
        super().__init__(learning_rate=learning_rate, dtype=dtype)
        self.rho = rho
        self.epsilon = epsilon
        self.decay = decay

    @classmethod
    def from_model(cls, model: "Model") -> "RMSProp":
        return RMSProp(learning_rate=model.learning_rate,
                       rho=model.optimizer_rho,
                       epsilon=model.optimizer_epsilon,
                       decay=model.optimizer_decay,
                       dtype=model.dtype)



import os
import copy
import numpy as np

from pydtnn.datasets.dataset import Dataset
from pydtnn.utils.tensor import TensorFormat
from pydtnn.utils.constants import ArrayShape
from pydtnn.utils import random

from typing import TYPE_CHECKING, override, Generator
if TYPE_CHECKING:
    from pydtnn.model import Model

type DataPath = str
type ClassName = int


class Folder(Dataset):
    """
    This class will receive the path to a dataset divided in different sub-folders where every sub-folder is a different data class, and will
    generate the samples.
    For example:
    - Dataset:
        - A: img1, img2
        - B: img3, img4, img5
        - C: img6

    The Dataset is composed by img1 and img2, which belongs to the class A; img3, img4 and img5, which belong to class the class B; and img6, which belongs to class C.
    """

    def __init__(self, model: "Model", force_test_as_validation=False, debug=False):
        """
        Args:
            model (Model): Model's object.
            force_test_as_validation (bool): True to force the use of the test dataset as validation. default: False.
            debug (bool): True to show debug prints. default: False.
        """

        # NOTE: Validation dataset is extracted from the Test one.
        self.model = model
        if not os.path.isdir(self.model.dataset_path):
            raise NotADirectoryError(f"{self.model.dataset_path!r} should be a directory.")
        dataset_train_path = os.path.join(self.model.dataset_path, "train")
        dataset_test_path = os.path.join(self.model.dataset_path, "test")

        # self.new_size = (new_size, new_size) if isinstance(new_size, int) else new_size
        self._nsamples = [0, 0, 0]  # train, val, test
        self.labels_and_images = dict[Dataset.Part, list[tuple[ClassName, DataPath]]]()

        self.labels_and_images[Dataset.Part.TRAIN], num_classes_train, self._nsamples[Dataset.Part.TRAIN] = self._get_dict_class_and_file(path=dataset_train_path)
        self.labels_and_images[Dataset.Part.TEST], num_classes_test, self._nsamples[Dataset.Part.TEST] = self._get_dict_class_and_file(path=dataset_test_path)

        if num_classes_train != num_classes_test:
            raise ValueError(f"The number of train classes ({num_classes_train}) must be the same as the number of test classes {num_classes_test}.")

        input_shape = (3, 10, 10)  # synthetic
        output_shape = (num_classes_train, )

        super().__init__(model=model, train_nsamples=self._nsamples[Dataset.Part.TRAIN],
                         test_nsamples=self._nsamples[Dataset.Part.TEST],
                         input_shape=input_shape, output_shape=output_shape,
                         force_test_as_validation=force_test_as_validation,
                         debug=debug)

        self.labels_and_images[Dataset.Part.VAL] = copy.copy(self.labels_and_images[Dataset.Part.TEST] if self.test_as_validation else self.labels_and_images[Dataset.Part.TRAIN])

    def _get_dict_class_and_file(self, path: str) -> tuple[list[tuple[ClassName, DataPath]], int, int]:
        dict_class_file = dict[ClassName, set[DataPath]]()
        num_images = 0
        list_dir = sorted(os.listdir(path))
        num_classes = len(list_dir)
        for class_name in range(num_classes):
            file = list_dir[class_name]
            path_folder = os.path.join(path, file)
            if os.path.isdir(path_folder):
                data_set = set(file for file in [os.path.join(path_folder, file) for file in sorted(os.listdir(path_folder))] if os.path.isfile(file))
                dict_class_file[class_name] = data_set
                num_images += len(data_set)

        if len(dict_class_file.values()) == 0:
            raise ValueError(f"There are no directories in \'{path}\'.")

        labels_and_images = [(class_name, path_image) for class_name, set_path_image in dict_class_file.items() for path_image in set_path_image]

        return (labels_and_images, num_classes, num_images)
    # ---

    def _prepare_label(self, label: int, num_classes: ArrayShape) -> np.ndarray:
        """Transform class numer into class mask (ndarray 1D unit8)"""
        np_label = np.zeros(shape=num_classes, dtype=np.uint8, order="C")
        np_label[label] = 1
        return np_label

    @override
    def _init_actual_data(self):
        if not self.model.transform_resize:
            raise ValueError("Model transform_resize must be enabled for dataset!")
    # ---

    @override
    def _actual_data_generator(self, part: Dataset.Part) -> Generator[tuple[np.ndarray, np.ndarray]]:
        offset = self._local_offset[part]
        nsamples = self._local_nsamples[part]
        labels_and_images = self.labels_and_images[part]

        if part is Dataset.Part.TRAIN and self.model.augment_shuffle:
            random.shuffle(labels_and_images)

        labels_and_images = labels_and_images[offset:offset + nsamples]

        for label, path_image in labels_and_images:
            x = self._load_rgb_image(path_image)
            y = self._prepare_label(label, self.output_shape)

            # Add N dimension
            x = x[None, ...]
            y = y[None, ...]

            # Set tensor format
            x = self.model.encode_tensor(x)

            # Set dtype and order
            x = x.astype(dtype=self.model.dtype, order="C")
            y = y.astype(dtype=self.model.dtype, order="C")

            # Inplace normalization
            x /= 255.0

            yield x, y



from pydtnn.backends.cpu.layers.abstract.block_layer import AbstractBlockLayerCPU
from pydtnn.layers.addition_block import AdditionBlock
from pydtnn.tracers.events import PYDTNN_MDL_EVENT, PYDTNN_MDL_EVENTS, PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, \
    PYDTNN_EVENT_FINISHED, PYDTNN_MDL_EVENT_enum, PYDTNN_OPS_EVENT_enum
import numpy as np


class AdditionBlockCPU(AdditionBlock[np.ndarray], AbstractBlockLayerCPU):

    def forward(self, x: np.ndarray) -> np.ndarray:

        num_paths = len(self.paths)
        p = self.paths[0]
        x_forward = x
        for layer in p:
            self.model.tracer.emit_event(PYDTNN_MDL_EVENT, layer.id * PYDTNN_MDL_EVENTS + PYDTNN_MDL_EVENT_enum.FORWARD)
            x_forward = layer.forward(x_forward)
            self.model.tracer.emit_event(PYDTNN_MDL_EVENT, PYDTNN_EVENT_FINISHED)
        sum_forwards = x_forward

        for i in range(1, num_paths):
            p = self.paths[i]
            x_forward = x
            for layer in p:
                self.model.tracer.emit_event(PYDTNN_MDL_EVENT, layer.id * PYDTNN_MDL_EVENTS + PYDTNN_MDL_EVENT_enum.FORWARD)
                x_forward = layer.forward(x_forward)
                self.model.tracer.emit_event(PYDTNN_MDL_EVENT, PYDTNN_EVENT_FINISHED)
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_ELTW_SUM)
            np.add(sum_forwards, x_forward, out=sum_forwards,
                   dtype=self.model.dtype, order="C")
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return np.asarray(sum_forwards, dtype=self.model.dtype, order='C', copy=None)

    def backward(self, dy: np.ndarray) -> np.ndarray:
        num_paths = len(self.paths)
        p = self.paths[0]
        dx_backward = dy
        for layer in reversed(p):
            self.model.tracer.emit_event(PYDTNN_MDL_EVENT, layer.id * PYDTNN_MDL_EVENTS + PYDTNN_MDL_EVENT_enum.FORWARD)
            dx_backward = layer.backward(dx_backward)
            self.model.tracer.emit_event(PYDTNN_MDL_EVENT, PYDTNN_EVENT_FINISHED)
        dx = dx_backward

        for i in range(1, num_paths):
            p = self.paths[i]
            dx_backward = dy
            for layer in reversed(p):
                self.model.tracer.emit_event(PYDTNN_MDL_EVENT, layer.id * PYDTNN_MDL_EVENTS + PYDTNN_MDL_EVENT_enum.BACKWARD)
                dx_backward = layer.backward(dx_backward)
                self.model.tracer.emit_event(PYDTNN_MDL_EVENT, PYDTNN_EVENT_FINISHED)

            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_ELTW_SUM)
            np.add(dx, dx_backward, out=dx,
                   dtype=self.model.dtype, order="C")
            self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return np.asarray(dx, dtype=self.model.dtype, order='C', copy=None)



from typing import TYPE_CHECKING

from numpy import ndarray

from pydtnn.schedulers.scheduler_with_loss_or_metric import SchedulerWithLossOrMetric

if TYPE_CHECKING:
    from pydtnn.model import Model


class StopAtLoss(SchedulerWithLossOrMetric):
    """
    StopAtLoss LRScheduler
    """

    def __init__(self, loss_or_metric: str = "", threshold_value=0, verbose=True):
        # NOTE: loss_or_metric default value is "val_accuracy" in Parser.
        super().__init__(loss_or_metric, verbose)
        self.threshold_value = threshold_value

    def on_epoch_end(self, train_loss: ndarray[float], val_loss: ndarray[float]) -> None:
        idx = self._get_idx()
        self.epoch_count += 1
        loss = val_loss if self.is_val_metric else train_loss
        if ("accuracy" in self.loss_or_metric and loss[idx] > self.threshold_value) or \
                ("accuracy" not in self.loss_or_metric and loss[idx] < self.threshold_value):
            self.stop_training = True
            self.log("Metric '{self.loss_or_metric}' reached threshold value {self.threshold_value}, stop training.")

    @classmethod
    def from_model(cls, model: "Model") -> "StopAtLoss":
        return StopAtLoss(model.stop_at_loss_metric,
                          model.stop_at_loss_threshold)



from typing import TYPE_CHECKING

import numpy as np

from pydtnn.datasets.custom_dataset import CustomDataset
from pydtnn.utils.tensor import TensorFormat
from pydtnn.utils.constants import ArrayShape

if TYPE_CHECKING:
    from pydtnn.model import Model


def archive(model: "Model", force_test_as_validation=False, debug=False) -> "CustomDataset":
    """
    Archived Dataset

    Load from a NPZ with x_train, y_train, x_test, y_test attributes.
    Train and Test must have matching types, shapes and dtypes.
    X must be in a NDArray with NCHW shape and float64 dtype.
    Y must be in a NDArray with N (or more) and float64 dtype.
    """

    with np.load(model.dataset_path) as data:
        data: dict[str, np.ndarray]
        x_train = data["x_train"]
        y_train = data["y_train"]
        x_test = data["x_test"]
        y_test = data["y_test"]
        input_shape: ArrayShape = x_train.shape[1:]

    # Ensure dataset is in model.tensor_format
    x_train = model.encode_tensor(x_train)
    x_test = model.encode_tensor(x_test)

    # Ensure dataset is in model.dtype
    match model.dtype:
        case np.float64:
            pass
        case np.float32:
            x_train, y_train = x_train.astype(np.float32), y_train.astype(np.float32)
            x_test, y_test = x_test.astype(np.float32), y_test.astype(np.float32)
        case _:
            raise NotImplementedError(f"Unsupported model dtype {model.dtype}")

    # Ensure dataset transformations are applied
    x_train, y_train = x_train.copy(), y_train.copy()
    x_test, y_test = x_test.copy(), y_test.copy()

    # Create dataset
    dataset = CustomDataset(
        model,
        x_train=x_train,
        y_train=y_train,
        x_test=x_test,
        y_test=y_test,
        input_shape=input_shape,
        force_test_as_validation=force_test_as_validation,
        debug=debug
    )

    # Debug information
    if dataset.debug:
        print(f"Import: {dataset.model.dataset_raw_path}")
        print(f"x_train: {x_train.shape}")
        print(f"y_train: {y_train.shape}")
        print(f"x_test: {x_test.shape}")
        print(f"y_test: {y_test.shape}")

    return dataset
